/* ---------------------------------------------------------------------- */
/*                        FULL SCREEN DEBUGGER                            */
/*                                                                        */
/* Copyright 1994 by Morten Welinder, terra@diku.dk                       */
/* ---------------------------------------------------------------------- */
/*

This debugger is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

This debugger is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with djgpp; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
/* ---------------------------------------------------------------------- */
#define FULLSCR_VERSION 1.04
#define FS_MAJOR_VER 1
#define FS_MINOR_VER 04

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <pc.h>
#include <dos.h>
#include <dpmi.h>
#include <go32.h>
#include <keys.h>
#include <sys/stat.h>
#include <fcntl.h>
#define K_BackTab 0x10f
#define K_ELeft 0x24b
#include <setjmp.h>
#define far
#include "ed.h"
#include "syms.h"
#include "paging.h"
#include "unassmbl.h"
#include "npx.h"
#include "serial.h"
#include "fsfuncts.h"
#include "xsupport.h"
#include "perform.h"
/* ---------------------------------------------------------------------- */
/* Information about the initialization state of the debugger.  Actually
   I'm not sure that longjmp should ever be permitted.  */
int can_longjmp = 0;
jmp_buf debugger_jmpbuf;

/* ---------------------------------------------------------------------- */
/* Information added to use com port - by Long Doan */
static int using_com_port_num;
static int first_transfer = 1;
static char *transfer_screen_save;
static char *my_line;
static char *small_line;
int is_in_running_mode = 0;   /* Communicate with the comm driver. */

/* Display information.  */
int debug_screen_p;
char *user_screen_save, *debug_screen_save;
unsigned char screen_attr;
unsigned char screen_attr_normal;
unsigned char screen_attr_source;
unsigned char screen_attr_focus;
unsigned char screen_attr_ffocus;
unsigned char screen_attr_break;
unsigned char screen_attr_message;
unsigned char screen_attr_error;
unsigned char screen_attr_resize;
unsigned char screen_attr_asm;
unsigned char screen_attr_help;
unsigned char screen_attr_addr;
int cols, max_cols;
int rows, max_rows;
int toplines, bottomlines;
int main_x1, main_x2, main_x3;


/* Information about panes.  */
#define PANECOUNT 11
static int pane, pane_positions[PANECOUNT], pane_pos;
static word32 data_dump_origin, data_dump_last, data_dump_size;
static word32 code_dump_origin, code_dump_last;
static word32 *code_pane_pos, *stack_dump_pos;
static int stack_dump_origin, stack_dump_last, stack_dump_more;
static int breakpoint_origin;
static char **whereis_pane_text;
static char **module_pane_text;
static char **help_pane_text;
static int whereis_text_count, whereis_origin;
static int module_text_count, module_origin;
static int help_text_count, help_origin;
static int code_pane_active;
static int npx_pane_active;
static int stack_pane_active;
static int info_pane_active;
static int whereis_pane_active;
static int module_pane_active;
static int help_pane_active;

/* Menus */
static MENU_ITEM File_Sub_Menu[] = { 
       "Current item   ", 1        , 0, 
       "Read/Write Log ", ITEM_TYPE, 0,
       "Resize Screen  ", ITEM_TYPE, 0,
       "Monochrome Mode", ITEM_TYPE, 0,
       "Color Mode     ", ITEM_TYPE, 0,
       "Redirect IO    ", ITEM_TYPE, 0,
       "Exit           ", ITEM_TYPE, 0,
       0 	        , 0        , 0
};
static MENU_ITEM Pane_Sub_Menu[] = {
       "Current   ", 1        , 0,
       "Previous  ", ITEM_TYPE, 0,
       "Next      ", ITEM_TYPE, 0,
       "Breakpoint", ITEM_TYPE, 0,
       "Code      ", ITEM_TYPE, 0,
       "Data      ", ITEM_TYPE, 0,
       "Flags     ", ITEM_TYPE, 0,
       "Help      ", ITEM_TYPE, 0,
       "Info      ", ITEM_TYPE, 0,
       "Module    ", ITEM_TYPE, 0,
       "Npx       ", ITEM_TYPE, 0,
       "Registers ", ITEM_TYPE, 0,
       "Stack     ", ITEM_TYPE, 0,
       "Where-is  ", ITEM_TYPE, 0,
       "Resize    ", ITEM_TYPE, 0,
       0 	   , 0        , 0
};
static MENU_ITEM Help_Sub_Menu[] = {
       "Current   ", 1        , 0,
       "About     ", ITEM_TYPE, 0,
       "Help      ", ITEM_TYPE, 0,
       0           , 0        , 0
};
static MENU_ITEM Misc_Sub_Menu[] = {
       "Current    ", 1        , 0,
       "User Screen", ITEM_TYPE, 0,
       "Evaluate   ", ITEM_TYPE, 0,
       "Refresh  ^Z", ITEM_TYPE, 0,
       0            , 0        , 0
};
static MENU_ITEM Main_Menu[] = {
       " Current "    , 1,         0,
       " File  "      , MENU_TYPE, File_Sub_Menu,
       " Pane  "      , MENU_TYPE, Pane_Sub_Menu,
       " Local Menu  ", ITEM_TYPE, 0,
       " Misc  "      , MENU_TYPE, Misc_Sub_Menu,
       " Help  "      , MENU_TYPE, Help_Sub_Menu,
       0 	    , 0        , 0
};                

/* Odds and ends.  */
#define MAXINSTLEN 16
static int first_step;
static word32 main_entry;
char *read_buffer;
NPX npx; /* non-static because declared so in "npx.h" */
static void redraw (int);
static char hexchars[] = "0123456789abcdef";
/* ---------------------------------------------------------------------- */
/* The presentation order of registers in the register pane.  The three
   tables must of course match.  */

static char *regs_names[] = {
  "eax", "ebx", "ecx", "edx", "esi", "edi", "ebp", "esp",
  "cs",  "ds",  "es",  "fs",  "gs",  "ss",
  "eip", "flg" };

static word32 *regs_addr[] = {
  &a_tss.tss_eax, &a_tss.tss_ebx, &a_tss.tss_ecx, &a_tss.tss_edx,
  &a_tss.tss_esi, &a_tss.tss_edi, &a_tss.tss_ebp, &a_tss.tss_esp,
  (word32 *)&a_tss.tss_cs, (word32 *)&a_tss.tss_ds,
  (word32 *)&a_tss.tss_es, (word32 *)&a_tss.tss_fs,
  (word32 *)&a_tss.tss_gs, (word32 *)&a_tss.tss_ss,
  &a_tss.tss_eip, &a_tss.tss_eflags };

/* g: general, !: special, s: segment, f: flags, \0: end-of-table */
static char regs_type[] = "ggggggg!ssssss!f";
/* ---------------------------------------------------------------------- */
/* Breakpoint data.  When breakpoints are actually put into the cpu debug
   registers, data breakpoints have first priority.  Any code breakpoints
   left over are set by patching the code to contain "Int 3".  */
typedef enum { BP_Code = 0, BP_Write = 1, BP_Read = 3 } BP_TYPE;
typedef enum { BP_Normal = 0, BP_Count = 1, BP_Condition = 2 } BP_CODE_TYPE;

typedef struct
{
  word32 addr;
  BP_TYPE type;
  BP_CODE_TYPE code_type;
  int count;
  char condition[80];
  unsigned char length;       /* 1, 2, or 4 bytes */
  unsigned char saved;
  unsigned char savedbyte;    /* Patched-away byte. */
} BP_ENTRY;

static int breakpoint_count;
static BP_ENTRY *breakpoint_table;
static unsigned char int03 = 0xcc;
/* ---------------------------------------------------------------------- */
/* Store the contents of the NPX in the global variable `npx'.  */
static void
save_npx (void)
{
asm(                          "\n\
	inb     $0xa0,%al      \n\
	testb   $0x20,%al      \n\
	jz      Lfclex_done    \n\
	xorl    %eax,%eax      \n\
	outb    %al,$0xf0      \n\
	movb    $0x20,%al      \n\
	outb    %al,$0xa0      \n\
	outb    %al,$0x20      \n\
Lfclex_done:                   \n\
	movl    $_npx, %eax    \n\
	fnsave  (%eax)         \n\
	fwait                  "
    );
}
/* ---------------------------------------------------------------------- */
/* Reload the contents of the NPX from the global variable `npx'.  */
static void
load_npx (void)
{
asm(                                                                   "\n\
	movl    $_npx, %eax                                             \n\
/*      movb    $0,4(%eax)  */          /* clear pending exceptions */  \n\
	frstor  (%eax)                                                  "
    );
/* At least on my 486 there is no need to clear pending exceptions.  */
}
/* ---------------------------------------------------------------------- */
static unsigned int
get_com(unsigned int msecs)
{
  int ch, stop = 0;
  struct time now;
  unsigned long now100, then100 = 0;

  if (msecs)
  { gettime (&now);
    then100
      = ((now.ti_hour * 60 + now.ti_min) * 60 + now.ti_sec) * 100
        + now.ti_hund + msecs;
  }

  do {
    ch = com_rx();
    if (ch)
      stop = 1;
    else 
      if (msecs)
      { gettime (&now);
        now100 = ((now.ti_hour * 60 + now.ti_min) * 60 + now.ti_sec) * 100
                 + now.ti_hund;
        if ((now100 < 100) || (now100 >= then100))
          stop = 1;
      } 
  } while (stop == 0);
  return ch;
}
/* ---------------------------------------------------------------------- */
static unsigned int key_buf[3];
static int key_buf_level = 0;

unsigned int
getykey (void)
{
  if (using_com_port_num)
  { unsigned int ch, ch1, ch2;

    if (key_buf_level)
    { ch = key_buf[0];
      key_buf[0] = key_buf[1];
      key_buf[1] = key_buf[2];
      key_buf_level--;
    }
    else
      ch = get_com(0);

    switch (ch)
    { case 0x1b:
	if (key_buf_level)
	{ ch1 = key_buf[0];
	  key_buf[0] = key_buf[1];
	  key_buf[1] = key_buf[2];
	  key_buf_level--;
	}
	else
	  ch1 = get_com(20);
	switch (ch1)
	{ case '[':
	    ch2 = get_com(0);
	    switch (ch2)
	    { case 'A':             /* ^[[A == Up       */
		return K_Up;
	      case 'B':             /* ^[[B == Down     */
		return K_Down;
	      case 'C':             /* ^[[C == Right    */
		return K_Right;
	      case 'D':             /* ^[[D == Left     */
		return K_Left;
	      case 'H':             /* ^[[H == Home     */
		return K_Home;
	      case 'K':             /* ^[[K == End      */
		return K_End;
	      default: 
		key_buf[key_buf_level++] = ch1;
		key_buf[key_buf_level++] = ch2;
	    }              
          case 0:
            return K_Escape;
	  default:
	    key_buf[key_buf_level++] = ch1;
	}
	return K_Escape;
      case 0x0d:
	return K_Return;
      case K_Control_Y:   /* 0x1e:  Ctrl-`  */
        return K_PageUp;  
      case K_Control_V:   /* 0x1f:  Ctrl--  */
        return K_PageDown;
      default:
        return ch;
    }
  }
  else
    return getxkey();
}
/* ---------------------------------------------------------------------- */
static void
ansi_goto (int row, int col)
{
  sprintf (small_line, "a[%d;%df", row + 1, col + 1);
  small_line[0] = 27;
}
/* ---------------------------------------------------------------------- */
static void
ansi_color (int bg, int fg)
{
  int is_bold = 0, is_blink = 0;

  if (bg > 7)
  { is_blink = 1;
    bg &= 7;
  }
  if (fg > 7)
  { is_bold = 1;
    fg &= 7;
  }
  switch (bg)
  { case A_blue:
      bg = A_red;
      break;
    case A_red:
      bg = A_blue;
      break;
    case A_brown:
      bg = A_cyan;
      break;
    case A_cyan:
      bg = A_brown;
      break;
  }
  switch (fg)
  { case A_blue:
      fg = A_red;
      break;
    case A_red:
      fg = A_blue;
      break;
    case A_brown:
      fg = A_cyan;
      break;
    case A_cyan:
      fg = A_brown;
      break;
  }
  bg += 40;
  fg += 30;
  if (is_bold)
    if (is_blink)
      sprintf(small_line, "a[0;1;5;%d;%dm",bg,fg);
    else
      sprintf(small_line, "a[0;1;%d;%dm",bg,fg);
  else
    if (is_blink)
      sprintf(small_line, "a[0;5;%d;%dm",bg,fg);
    else
      sprintf(small_line, "a[0;%d;%dm",bg,fg);
  small_line[0] = 27;
}
/* ---------------------------------------------------------------------- */
static void
screen_goto (int row, int col)
{  
  ansi_goto (row, col);
  com_txs (small_line);
}
/* ---------------------------------------------------------------------- */
static void
screen_scroll (int lines)
{  
  int i;

  screen_goto (rows, 0);
  for (i = 0; i < lines; i++)
    com_tx (10);
  screen_goto (0, 0);
}
/* ---------------------------------------------------------------------- */
inline static void
put (int x, int y, unsigned char *txt)
{
  unsigned char *p
    = (unsigned char *)(debug_screen_save + 3) + 2 * (max_cols * y + x);
  while (*txt)
    *p++ = *txt++,
    *p++ = screen_attr;
}
/* ---------------------------------------------------------------------- */
inline void
putl (int x, int y, int l, unsigned char *txt)
{
  unsigned char *p
    = (unsigned char *)(debug_screen_save + 3) + 2 * (max_cols * y + x);
  while (*txt && l > 0)
    *p++ = *txt++,
    *p++ = screen_attr,
    l--;
  while (l-- > 0)
    *p++ = ' ',
    *p++ = screen_attr;
}
/* ---------------------------------------------------------------------- */
inline static void
clear_screen (void)
{
  int l = rows * max_cols;

  unsigned char *p
    = (unsigned char *)(debug_screen_save + 3);
  while (l-- > 0)
    *p++ = ' ',
    *p++ = screen_attr;
}
/* ---------------------------------------------------------------------- */
inline static void
draw (int x, int y, unsigned char ch, int delta, int count)
{
  short unsigned *p
    = (short unsigned *)(debug_screen_save + 3) + (max_cols * y + x);
  short unsigned attrch = ((unsigned)screen_attr << 8) + ch;
  while (count--)
    *p = attrch,
    p += delta;
}
/* ---------------------------------------------------------------------- */
static void
highlight (int x, int y, int len)
{
  unsigned short *p
    = (unsigned short *)(debug_screen_save + 4) + (max_cols * y + x);
  while (len--)
    *(unsigned char *)p = screen_attr,
    p++;
}
/* ---------------------------------------------------------------------- */
void
frame (int x1, int y1, int x2, int y2)
{
  if (!using_com_port_num)
  { draw (x1 + 1, y1, '', 1,    x2 - x1 - 1);
    draw (x1 + 1, y2, '', 1,    x2 - x1 - 1);
    draw (x1, y1 + 1, '', max_cols, y2 - y1 - 1);
    draw (x2, y1 + 1, '', max_cols, y2 - y1 - 1);
    put (x1, y1, "");
    put (x2, y1, "");
    put (x1, y2, "");
    put (x2, y2, "");
  } else
  { draw (x1 + 1, y1, '-', 1,    x2 - x1 - 1);
    draw (x1 + 1, y2, '-', 1,    x2 - x1 - 1);
    draw (x1, y1 + 1, '|', max_cols, y2 - y1 - 1);
    draw (x2, y1 + 1, '|', max_cols, y2 - y1 - 1);
    put (x1, y1, "+");
    put (x2, y1, "+");
    put (x1, y2, "+");
    put (x2, y2, "+");
  }
}
/* ---------------------------------------------------------------------- */
void
double_frame (int x1, int y1, int x2, int y2)
{
  if (!using_com_port_num)
  { draw (x1 + 1, y1, '', 1,    x2 - x1 - 1);
    draw (x1 + 1, y2, '', 1,    x2 - x1 - 1);
    draw (x1, y1 + 1, '', max_cols, y2 - y1 - 1);
    draw (x2, y1 + 1, '', max_cols, y2 - y1 - 1);
    put (x1, y1, "");
    put (x2, y1, "");
    put (x1, y2, "");
    put (x2, y2, "");
  } else
  { draw (x1 + 1, y1, '=', 1,    x2 - x1 - 1);
    draw (x1 + 1, y2, '=', 1,    x2 - x1 - 1);
    draw (x1, y1 + 1, '[', max_cols, y2 - y1 - 1);
    draw (x2, y1 + 1, ']', max_cols, y2 - y1 - 1);
    put (x1, y1, "*");
    put (x2, y1, "*");
    put (x1, y2, "*");
    put (x2, y2, "*");
  }
}
/* ---------------------------------------------------------------------- */
static void
get_line (char *screen, int line_num)
{
  int i;
  char *p2;
  unsigned char *p =
    (unsigned char *)(screen + 3) + 2 * max_cols * line_num;

  p2 = my_line;
  i = cols * 2;
  while (i)
  {
    *p2++ = *p++;
    i--;
  }        
  *p2++ = 0;
}
/* ---------------------------------------------------------------------- */
static void
put_line (char *screen, char *full_line, int line_num)
{
  unsigned char *p =
    (unsigned char *)(screen + 3) + 2 * max_cols * line_num;
  char *oneline = full_line;
  int i = cols * 2;

  while (i)
  {
    *p++ = *oneline++;
    i--;
  }
}
/* ---------------------------------------------------------------------- */
static void
get_differ (char *first_line, char *second_line)
{
  char *line_one = first_line, *line_two = second_line;
  char ch1_attr, ch1_ascii, ch2_attr, ch2_ascii;
  int len, i = 0;

  len = 1;
  my_line[0] = '\0';

  while (i < cols)
  { ch1_ascii = *line_one++;
    ch1_attr  = *line_one++;
    ch2_ascii = *line_two++;
    ch2_attr  = *line_two++;
    i++;

    if ((ch1_ascii != ch2_ascii) || (ch1_attr != ch2_attr))
    { len += 5;
      my_line[len - 6] = 255;
      my_line[len - 5] = 255;
      my_line[len - 4] = (unsigned char)i;
      my_line[len - 3] = ch2_ascii;
      my_line[len - 2] = ch2_attr;
      my_line[len - 1] = '\0';
      while (i < cols)
      {	ch1_ascii = *line_one++;
	ch1_attr  = *line_one++;
	ch2_ascii = *line_two++;
	ch2_attr  = *line_two++;
	i++;

	if ((ch1_ascii == ch2_ascii) && (ch1_attr == ch2_attr))
	  break;

	len += 2;
	my_line[len - 3] = ch2_ascii;
	my_line[len - 2] = ch2_attr;
	my_line[len - 1] = '\0';
      }
    }
  }
}
/* ---------------------------------------------------------------------- */
static int
com_filter (int old_char)
{
  switch (old_char)
  { case 0:            /* Cannot display  */          
    case 7:            /* Bell            */
    case 8:            /* Backspace       */
    case 10:           /* Lf              */
    case 13:           /* Cr              */
    case 127:          /* DEL             */
      return (int)('.');
    case 30:           /* Big Up Arrow    */
      return (int)('^');
    case 31:           /* Big Down Arrow  */
      return (int)('v');
    default:
      return old_char;
  }
}
/* ---------------------------------------------------------------------- */
static void
line_to_ansi (char *full_line, int line_num)
{
  int i = cols, len;
  unsigned char old_attr = 0, ch_attr, ch_ascii;
  char *oneline = full_line;

  ansi_goto (line_num, 0);           
  sprintf (my_line, small_line);
  len = strlen (small_line);

  while (i)
  { ch_ascii = *oneline++;
    ch_attr  = *oneline++;

    if (old_attr != ch_attr)
    { old_attr = ch_attr;
      ansi_color (ch_attr >> 4, ch_attr & 0x0f);
      sprintf ((char *)(my_line + len), small_line);
      len += strlen (small_line);
    }
    my_line[len] = com_filter (ch_ascii);
    my_line[++len] = '\0';
    i--;
  }
}
/* ---------------------------------------------------------------------- */
static void
diff_to_ansi (unsigned char *diff_line, int line_num)
{
  char *oneline = diff_line;
  int i = 0, len = 0, dif_len = strlen(oneline);
  char old_attr = 0, ch_attr = 0, ch_ascii = 0;
  int is_pos = 0, pos = 0;
  unsigned char char_one, char_two;

  while (i < dif_len)
  { char_one = *oneline++;
    char_two = *oneline++;
    i += 2;
    if ((char_one == 255) && (char_two == 255))
    { pos = *oneline;
      if (pos <= cols)
      {	i++;
	oneline++;
	is_pos = 1;
	pos--;
      }
      else
      { ch_ascii = char_one;
        ch_attr  = char_two;
      }
    }
    else
    { ch_ascii = char_one;
      ch_attr  = char_two;
    }

    if (is_pos)
    { ansi_goto (line_num, pos);
      sprintf ((char *)(my_line + len), small_line); 
      len += strlen (small_line);
      is_pos = 0;
    }
    else
    { if (old_attr != ch_attr)
      {	old_attr = ch_attr;
	ansi_color (ch_attr >> 4, ch_attr & 0x0f);
        sprintf ((char *)(my_line + len), small_line);
        len += strlen (small_line);
      }
      my_line[len] = com_filter(ch_ascii);
      my_line[++len] = '\0';
    }
  }
}
/* ---------------------------------------------------------------------- */
static char new_line[1024] , old_line[1024];
static char diff_line[1024], ansi_line[1024];
/* ---------------------------------------------------------------------- */
static void
transfer_screen (char *screen)
{
  int line_num;         

  if (first_transfer)
  { screen_scroll (rows);
    line_num = 0;
    do 
    { get_line(screen, line_num);
      put_line(transfer_screen_save, my_line, line_num);
      strcpy (new_line, my_line);
      line_to_ansi (new_line , line_num);
      strcpy(ansi_line, my_line);
      com_txs (ansi_line);
    } while (++line_num < rows);
    first_transfer = 0;
    return;
  }

  line_num = 0;
  do 
  { get_line(screen, line_num);
    strcpy(new_line, my_line);
    get_line(transfer_screen_save, line_num);
    strcpy(old_line, my_line);

    if (strncmp (new_line, old_line, cols * 2))
    { put_line(transfer_screen_save, new_line, line_num);
      get_differ (old_line, new_line);
      strcpy(diff_line, my_line);
      diff_to_ansi (diff_line, line_num);
      strcpy(ansi_line, my_line);
      line_to_ansi (new_line , line_num);
      strcpy(diff_line, my_line);
      if ((strlen (ansi_line) < strlen (diff_line)))
	com_txs (ansi_line);
      else
	com_txs (diff_line);
    }

    line_num++;
  } while (line_num < rows);
}
/* ---------------------------------------------------------------------- */
static void
recieve_screen (unsigned char *screen)
{
  unsigned char *p2 = (unsigned char *)(screen + 3);
  unsigned char *p  = (unsigned char *)(transfer_screen_save + 3);
  int i = max_rows * max_cols * 2;

  while (i)
  {
    *p2++ = *p++;
    i--;
  }
}
/* ---------------------------------------------------------------------- */
void
put_screen (char *screen)
{
  if (using_com_port_num)
    transfer_screen(screen);
  else
    switch (*screen++)
      {
      case 0:
	/* Text screen.  */
	ScreenSetCursor (screen[0], screen[1]);
	ScreenUpdate (screen + 2);
	break;
      }
}
/* ---------------------------------------------------------------------- */
char *
get_screen ()
{
  char *p;
  int r, c;

  p = malloc (max_cols * max_rows * 2 + 3);    /* Was cols * rows * ... */
  if (using_com_port_num)
    {
      if (transfer_screen_save)
	recieve_screen (p);
      else
        for (r = 0; r < max_rows; r++)
          for (c = 0; c < max_cols; c++)
          { p[2 * (r * max_cols + c) + 3] = ' ';
            p[2 * (r * max_cols + c) + 4] = screen_attr;
          }
    }
  else
    {
      ScreenGetCursor (&r, &c);
      p[0] = 0;
      p[1] = r;
      p[2] = c;
      ScreenRetrieve (p + 3);
    }
  return p;
}
/* ---------------------------------------------------------------------- */
/* Display the debugger screen (if not already displayed).  */
static void
debug_screen ()
{
  if (!debug_screen_p)
    {
      user_screen_save = get_screen ();
      if (using_com_port_num)
        screen_scroll (rows);
      put_screen (debug_screen_save);
      debug_screen_p = 1;
    }
}
/* ---------------------------------------------------------------------- */
/* Display the user screen (if not already displayed).  */
static void
user_screen ()
{
  if (debug_screen_p)
    {
      if (using_com_port_num)
        screen_scroll (rows);
      else
        put_screen (user_screen_save);
      free (user_screen_save);
      debug_screen_p = 0;
    }
}
/* ---------------------------------------------------------------------- */
/* Reportedly, `sleep' & `gettimeofday' are buggy under Dpmi -- emulate.  */
static int
mysleep (int secs)
{
  struct time now;
  unsigned now100, then100;

  gettime (&now);
  then100
    = ((now.ti_hour * 60 + now.ti_min) * 60 + now.ti_sec + secs) * 100
      + now.ti_hund;
  do
    {
      gettime (&now);
      now100 = ((now.ti_hour * 60 + now.ti_min) * 60 + now.ti_sec) * 100
	+ now.ti_hund;
      if (now100 < 100)
	break; /* Day turnover */
      if (using_com_port_num)
      { if (get_com (0))
          break;
      }
      else
        if (bioskey (1))
          break;
    }
  while (now100 < then100);
  return 0;
}
/* ---------------------------------------------------------------------- */
static int
short_sleep (unsigned long msecs)
{
  struct time now;
  unsigned long now100, then100;

  gettime (&now);
  then100
    = ((now.ti_hour * 60 + now.ti_min) * 60 + now.ti_sec) * 100
      + now.ti_hund + msecs;
  do
    {
      gettime (&now);
      now100 = ((now.ti_hour * 60 + now.ti_min) * 60 + now.ti_sec) * 100
	+ now.ti_hund;
      if (now100 < 100 || bioskey (1))
	break; /* Day turnover */
    }
  while (now100 < then100);
  if (bioskey(1))
    return 1;
  return 0;
}
/* ---------------------------------------------------------------------- */
/* Display a message in CLASS using FMT and ... as printf parameters.
   The class determines the colour of the message and for how long time
   it is displayed.  */

typedef enum { CL_Info, CL_Msg, CL_Error } CL_TYPE;

static void
message (CL_TYPE class, char *fmt, ...)
{
  char *save, *buf = alloca (cols);
  unsigned char saveattr = screen_attr;
  int len, y = rows / 2;

  vsprintf (buf, fmt, (&fmt) + 1);
  len = strlen (buf);
  save = debug_screen_save;
  debug_screen_save = get_screen ();
  switch (class)
    {
    case CL_Error:
      screen_attr = screen_attr_error;
      break;
    default:
      screen_attr = screen_attr_message;
    }
  frame (1, y - 1, cols - 2, y + 1);
  draw (2, y, ' ', 1, cols - 4);
  put ((cols - len) / 2, y, buf);
  screen_attr = saveattr;
  put_screen (debug_screen_save);
  switch (class)
    {
    case CL_Info:
      mysleep (2);
      break;
    case CL_Msg:
    case CL_Error:
      (void) getykey ();
    }
  free (debug_screen_save);
  debug_screen_save = save;
  put_screen (debug_screen_save); 
  free (buf);
}
/* ---------------------------------------------------------------------- */
/* Read a string from the keyboard to `read_buffer'.  The entry is started
   with STARTTEXT.  */
int
read_string (char *starttext)
{
  char *save;
  int key, esc = 0, pos, leave = 0, y = rows / 2;

  save = debug_screen_save;
  debug_screen_save = get_screen ();
  strcpy (read_buffer, starttext);
  pos = strlen (read_buffer);

  frame (1, y - 1, cols - 2, y + 1);
  do
    {
      draw (2, y, ' ', 1, cols - 4);
      put (3, y, read_buffer);
      pos = strlen (read_buffer);
      put_screen (debug_screen_save);
      if (!using_com_port_num)
	ScreenSetCursor (y, 3 + pos);
      key = getykey ();
      switch (key)
	{
	case K_Return:
	  leave = 1;
	  esc = 0;
	  break;
	case K_Escape:
	  leave = 1;
	  esc = 1;
	  read_buffer[0] = '\0';
	  break;
	case K_BackSpace:
	  if (pos > 0)
	    read_buffer[--pos] = '\0';
	  break;
	default:
	  if (key >= ' ' && key <= 0xff && pos <= cols - 7)
	    {
	      read_buffer[pos++] = key & 0xff;
	      read_buffer[pos] = '\0';
	    }
	}
    }
  while (!leave);
  free (debug_screen_save);
  debug_screen_save = save;
  put_screen (debug_screen_save);
  return esc;
}
/* ---------------------------------------------------------------------- */
/* Read a string from the keyboard to `read_buffer'.  */
int
my_read_string (char *prompt)
{
  char *save;
  int key, esc = 0, pos, leave = 0, y = rows / 2;
  int prompt_len = strlen(prompt);

  save = debug_screen_save;
  debug_screen_save = get_screen ();
  *read_buffer = '\0';
  screen_attr = screen_attr_focus;

  frame (1, y - 1, cols - 2, y + 1);
  do
    {
      draw (2, y, ' ', 1, cols - 4);
      put (3, y, prompt);
      put (3 + prompt_len, y, read_buffer);
      pos = strlen (read_buffer) + prompt_len;
      put_screen (debug_screen_save);
      if (!using_com_port_num)
	ScreenSetCursor (y, 3 + pos);
      key = getykey ();
      switch (key)
	{
	case K_Return:
	  leave = 1;
	  esc = 0;
	  break;
	case K_Escape:
	  leave = 1;
	  esc = 1;
	  read_buffer[0] = '\0';
	  break;
	case K_BackSpace:
	  if (pos > prompt_len)
	  {
	    read_buffer[pos - prompt_len - 1] = '\0';
	    pos--;
	  }
	  break;
	default:
	  if (key >= ' ' && key <= 0xff && pos <= cols - 7)
	    {
	      read_buffer[pos-prompt_len] = key & 0xff;
	      pos++;
	      read_buffer[pos-prompt_len] = '\0';
	    }
	}
    }
  while (!leave);
  free (debug_screen_save);
  debug_screen_save = save;
  put_screen (debug_screen_save);
  return esc;
}
/* ---------------------------------------------------------------------- */
typedef enum { not, com, mul, divi, add, sub, shl, shr,
	       and, xor, or, land, lor } OPR_Name;
typedef enum { equal, equal2, great_equal, less_equal, not_equal,
	       greater, less } SEP_Name;
typedef struct {
	char *str;
	OPR_Name name;
} OPR_TYPE;
typedef struct {
	char *str;
	SEP_Name name;
} SEP_TYPE;

static OPR_TYPE opers[] = {
       "||", lor,     /* Lowest priority first */
       "&&", land,
       "|" , or,
       "^" , xor,
       "&" , and,
       ">>", shr,
       "<<", shl,
       "-" , sub,
       "+" , add,
       "/" , divi,
       "*" , mul,
       "~" , com,     /* Unary, left until the end. */
       "!" , not,
       0   ,0
};
static SEP_TYPE sepers[] = {
       "==", equal,
       ">=", great_equal,
       "<=", less_equal,
       "<>", not_equal,
       "!=", not_equal,
       "=" , equal2,
       ">" , greater,
       "<" , less,
       0   , 0
};
static char l_str[140], r_str[140];

void
break_up (char *expr, int *opr_num, int *found)
{
  int i, len = strlen(expr);
  char *ptr = 0;

  for (i = 0; i < 140; i++)
    l_str[i] = '\0';
  for (i = 0; i < 140; i++)
    r_str[i] = '\0';
  *found = 0;
  for (i = 0; !(*found) && opers[i].str; i++)
  {
    ptr = strstr(expr, opers[i].str);
    *found = (unsigned int)ptr;
  }
  if (*found)
  {
    i--;
    strncpy(l_str, expr, len - strlen(ptr));
    ptr++;
    switch (opers[i].name)
    {
      case shl:
      case shr:
      case land:
      case lor:
	ptr++;
    }
    strcpy(r_str, ptr);
    *opr_num = i;
  }
}
/* ---------------------------------------------------------------------- */
long
parse_expression (int size, char *expr, int *okp)
{
  long l_val, r_val, result = 0;
  char *l_expr, *r_expr;
  int opr_num;
  int found, ok;

  break_up(expr, &opr_num, &found);
  if (found)
  { l_expr = strdup(l_str);
    r_expr = strdup(r_str);
    r_val = parse_expression(size, r_expr, &ok);
    free (r_expr);
    if (ok == 0)
    { free (l_expr);
      *okp = 0;
      return 0;
    }
    l_val = parse_expression(size, l_expr, &ok);
    free (l_expr);
    if (ok == 0)
    { *okp = 0;
      return 0;
    }
    switch (opers[opr_num].name)
    { case not:
	result = !r_val;
	break;
      case com:
	result = ~r_val;
	break;
      case mul:
	result = l_val * r_val;
	break;
      case divi:
	if (r_val == 0)
	{ *okp = 0;
	  return 0;
	}
	else
	  result = l_val / r_val;
	break;
      case add:
	result = l_val + r_val;
	break;
      case sub:
	result = l_val - r_val;
	break;
      case shl:
	result = l_val << r_val;
	break;
      case shr:
	result = l_val >> r_val;
	break;
      case and:
	result = l_val & r_val;
	break;
      case xor:
	result = l_val ^ r_val;
	break;
      case or :
	result = l_val | r_val;
	break;
      case land:
	result = l_val && r_val;
	break;
      case lor:
	result = l_val || r_val;
	break;
    }
  } else
  { result = syms_name2val(expr);
    if (undefined_symbol)
    {
      *okp = 0;
      return 0;
    }
    *okp = 1;
  }
  switch (size)
  { case 1:
      return (short)(result & 0xff);
    case 2:
      return (int)(result & 0xffff);
    case 4:          /* Fall through */
    default:
      return result;
  }
}
/* ---------------------------------------------------------------------- */
static void
seperate_expr (char *expr, int *size, int *sep_which)
{
  int stop = 0, i = 0, sep_len;
  char *sep_pos = 0, *my_expr;

  for (i = 0; i < 140; i++)
    l_str[i] = '\0';
  for (i = 0; i < 140; i++)
    r_str[i] = '\0';

  my_expr = expr;
  if ((my_expr[0] == '(') && (my_expr[2] == ')'))
  { my_expr += 3;
    switch (expr[1])
    { case '1':
	*size = 1;
	break;
      case '2':
	*size = 2;
	break;
      case '4':
	*size = 4;
	break;
      default:
	*size = 4;
    }
  }

  i = 0;
  while (!stop && sepers[i].str)
  { sep_pos = strstr(my_expr, sepers[i].str);
    stop = (unsigned int)sep_pos;
    i++;
  }

  if (stop)   /* There is some kind of seperator */
  { i--;
    sep_len = strlen(sep_pos);
    strncpy(l_str, my_expr, strlen(my_expr) - sep_len);
    sep_pos++;
    switch (sepers[i].name)
    { case equal:
      case great_equal:
      case less_equal:
      case not_equal:
	sep_pos++;
    }
    strcpy(r_str,sep_pos);
    *sep_which = i;
  } else
  { *sep_which = -1;
    strcpy(l_str, my_expr);
  }
}
/* ---------------------------------------------------------------------- */
static int
eval (char *expr, int *okp)
{ 
  word32 sym;

  sym = syms_name2val (expr);
  if (undefined_symbol)
    { message (CL_Error, "Expression not understood");
      return *okp = 0;
    }
  else
    { *okp = 1;
      return (int) sym;
    }
}
/* ---------------------------------------------------------------------- */
static long
read_eval (int *okp, char *starttext)
{
  long result;

  *okp = !read_string (starttext);
  if (*okp && read_buffer[0] != '\0')
  {
/*    return eval (read_buffer, okp);  */
    result = parse_expression (4, read_buffer, okp);
    if (!okp)
      message (CL_Error, "Expression not understood");
    return result;
  }
  else
    return *okp = 0;
}
/* ---------------------------------------------------------------------- */
int
valid_addr (word32 vaddr, int len)
{
  int a;

  if (vaddr >= 0xffffffffl - len)
    return 0;
  len--;
  for (a = 0; a < MAX_AREA; a++)
    if ((vaddr + len <= areas[a].last_addr) && (vaddr >= areas[a].first_addr))
      return 1;
  if (vaddr >= DOS_LOW_MEM_LO && vaddr <= DOS_LOW_MEM_HI)
    return 1;
  return 0;
}
/* ---------------------------------------------------------------------- */
inline static int
valid_instaddr (word32 vaddr)
{
  return valid_addr (vaddr, MAXINSTLEN);
}
/* ---------------------------------------------------------------------- */
/* Set physical breakpoint registers from virtual ones.  */
void
activate_breakpoints (void)
{
  int b, no;
  BP_ENTRY *bep;

  no = 0;
  edi.dr[7] = 0;
  /* First handle data breakpoints.  */
  for (b = 0, bep = breakpoint_table; b < breakpoint_count; b++, bep++)
    if (no <= 3 && bep->type != BP_Code)
      {
	bep->saved = 0;
	edi.dr[7] |= ((bep->type + ((bep->length - 1) << 2)) << (16 + 4 * no)
		      | (2 << (2 * no)));
	edi.dr[no] = bep->addr + edi.app_base;
	no++;
      }

  /* Now handle code breakpoint.  */
  for (b = 0, bep = breakpoint_table; b < breakpoint_count; b++, bep++)
    if (bep->type == BP_Code)
      if (no <= 3)
	{
	  bep->saved = 0;
	  edi.dr[7] |= ((BP_Code << (16 + 4 * no)) | (2 << (2 * no)));
	  edi.dr[no] = bep->addr + edi.app_base;
	  no++;
	  edi.dr[7] |= 0x00000300L;  /* For 386s we set GE & LE bits.  */
	}
      else
	{
	  bep->saved = valid_addr (bep->addr, 1);
	  if (bep->saved)
	    {
	      read_child (bep->addr, &bep->savedbyte, 1);
	      write_child (bep->addr, &int03, 1);
	    }
	}
}
/* ---------------------------------------------------------------------- */
/* Un-patch code.  */
void
deactivate_breakpoints (void)
{
  int b;
  BP_ENTRY *bep;

  for (b = 0, bep = breakpoint_table; b < breakpoint_count; b++, bep++)
    if (bep->saved)
      write_child (bep->addr, &bep->savedbyte, 1);
}
/* ---------------------------------------------------------------------- */
int
get_breakpoint (BP_TYPE type, int length, word32 addr)
{
  int b;

  for (b = 0; b < breakpoint_count; b++)
    if (breakpoint_table[b].type == type
	&& breakpoint_table[b].length == length
	&& breakpoint_table[b].addr == addr)
      return b;
  return -1;
}
/* ---------------------------------------------------------------------- */
static void
reset_breakpoint (int b)
{
  breakpoint_table[b] = breakpoint_table[--breakpoint_count];
  breakpoint_table
    = realloc (breakpoint_table, breakpoint_count * sizeof (BP_ENTRY));
}
/* ---------------------------------------------------------------------- */
int
set_breakpoint (BP_TYPE type, int length, word32 addr)
{
  int b;

  b = breakpoint_count;
  breakpoint_table
    = realloc (breakpoint_table, ++breakpoint_count * sizeof (BP_ENTRY));
  breakpoint_table[b].addr = addr;
  breakpoint_table[b].type = type;
  breakpoint_table[b].length = length;
  breakpoint_table[b].code_type = 0;
  breakpoint_table[b].count = 0;
  breakpoint_table[b].condition[0] = '\0';
  return b;
}
/* ---------------------------------------------------------------------- */
static void
edit_breakpoint (int bp_num)
{
  char *save, *name, *endp, *p = alloca(cols);
  int key, pos = 1, leave = 0;
  int x1 = 5, x2 = cols - x1 - 1;
  int y1 = (rows / 2) - 5, y2 = rows - y1 - 1;
  int i,len = 0, x_start, y_start, width = x2 - x1 - 13;
  int32 delta;
  word32 addr;
  int count, code_type;
  char condition[80];

  save = debug_screen_save;
  debug_screen_save = get_screen ();

  addr      = breakpoint_table[bp_num].addr;
  code_type = breakpoint_table[bp_num].code_type;
  count     = breakpoint_table[bp_num].count;
  strcpy(condition,breakpoint_table[bp_num].condition);

  screen_attr = screen_attr_ffocus;
  double_frame (x1, y1, x2, y2);
  screen_attr = screen_attr_normal;
  for (i = y1 + 1; i < y2; i++)
    putl(x1 + 1,i,x2 - x1 - 2," ");
  do
    {
      screen_attr = screen_attr_normal;

      name = syms_val2name (addr, &delta);
      if (name[0] != '0')
	if (delta && strlen (name) < width)
	  sprintf (p, "Address  : %s+%#lx", name, delta);
	else
	  sprintf (p, "Address  : %-*s", width, name);
      else
	sprintf(p, "Address  : %08lx",addr);

      putl (x1 + 1, y1 + 2, width + 11, p);
      sprintf(p, "Type     : %02d (%s)",
		 code_type,
		 (code_type == 0 ? "Normal" :
		   (code_type == 1 ? "Count down" : "Conditional")));
      putl (x1 + 1, y1 + 3, width + 11, p);
      sprintf(p, "Count    : %05d", count);
      putl (x1 + 1, y1 + 4, width + 11, p);
      sprintf(p, "Condition: %s",condition);
      putl (x1 + 1, y1 + 5, width + 11, p);
      screen_attr = (A_green << 4) + A_black;
      sprintf(p, "%s", "   Ok   ");
      putl (cols / 2 - 10, y1 + 7, 8, p);
      sprintf(p, "%s", " Cancel ");
      putl (cols / 2 + 1, y1 + 7, 8, p);
      screen_attr = (A_green << 4) + A_yellow;
      x_start = x1 + 12;
      y_start = y1 + pos + 1;
      switch (pos)
      {
	case 1:
	  len = width;
	  name = syms_val2name (addr, &delta);
	  if (name[0] != '0')
	    if (delta && strlen (name) < width)
	      sprintf (p, "%s+%#lx", name, delta);
	    else
	      sprintf (p, "%-*s", width, name);
	  else
	    sprintf(p,"%08lx",addr);
	  break;
	case 2:
	  len = 2;
	  sprintf(p,"%02d",code_type);
	  break;
	case 3:
	  len = 5;
	  sprintf(p,"%05d",count);
	  break;
	case 4:
	  len = width;
	  sprintf(p,"%s",condition);
	  break;
	case 5:
	  len = 8;
	  sprintf(p,"%s","   Ok   ");
	  x_start = cols / 2 - 10;
	  y_start++;
	  break;
	case 6:
	  len = 8;
	  sprintf(p,"%s"," Cancel ");
	  x_start = cols / 2 + 1;
	  break;
      }
      putl (x_start, y_start, len, p);
      put_screen (debug_screen_save);
      if (!using_com_port_num)
	ScreenSetCursor (y_start, x_start);
      key = getykey();
      switch(key)
      {
	case K_Up:
	case K_EUp:
	case K_BackTab:
	  pos--;
	  if (!pos)
	    pos = 6;
	  break;
	case K_Down:
	case K_EDown:
	case K_Tab:
	  pos++;
	  if (pos > 6)
	    pos = 1;
	  break;
	case K_Return:
	  switch (pos)
	  {
	    case 1:
	      if (!my_read_string("Enter new address: "))
		addr = syms_name2val(read_buffer);
	      break;
	    case 2:
	      if (!my_read_string("Enter new type: "))
	      {
		code_type = strtoul(read_buffer, &endp, 10);
		if (code_type > 2)
		  code_type = 0;
	      }
	      break;
	    case 3:
	      if (!my_read_string("Enter new count: "))
		count = strtoul(read_buffer, &endp, 10);
	      break;
	    case 4:
	      if (!my_read_string("Enter new condition: "))
		strcpy(condition,read_buffer);
	      break;
	    case 5:
	      leave = 1;
	      breakpoint_table[bp_num].addr      = addr;
	      breakpoint_table[bp_num].code_type = code_type;
	      breakpoint_table[bp_num].count     = count;
	      strcpy(breakpoint_table[bp_num].condition,condition);
	      break;
	    case 6:
	      leave = 1;
	      break;
	  }
	  break;
	case K_Escape:
	  leave = 1;
	  break;
      }
    }
  while (!leave);
  free (debug_screen_save);
  free (p);
  debug_screen_save = save;      
  put_screen (debug_screen_save);
}
/* ---------------------------------------------------------------------- */
/* From ORIGIN skip COUNT instructions forward (positive count) or
   backwards (negative count).  Backwards skipping cannot be assumed to
   be working 100% perfectly, but in the presense of debug information
   it is probably very, very close.

   This function works poorly close to the extremes of segments.  */

static int
code_skip (int origin, int count)
{
  int len, *next, i, j, k, instcount, done, leave;
  char *state, *inst, *source;

  if (count >= 0)
    {
      while (count-- > 0)
	if (valid_instaddr (origin))
	  {
	    (void) unassemble_proper (origin, &len);
	    origin += len;
	  }
	else
	  origin ++;
      return origin;
    }
  else
    {
      count = -count;
      instcount = MAXINSTLEN * (count + 16) + 1;
      next = alloca (instcount * sizeof (int));
      memset (next, 0, instcount * sizeof (int));
      state = alloca (instcount * sizeof (char));
      memset (state, 0, instcount * sizeof (char));

      done = 0;
      do
	{
	  for (i = 0; i < 2 * count; i++)
	    {
	      done++;
	      j = origin - done;
	      if (valid_instaddr (j))
		{
		  inst = unassemble_proper (j, &len);
		  source = unassemble_source (j);
		  next[done] = j + len;
		  if (source)
		    {
		      leave = 0;
		      k = done;
		      while (j < origin && state[k] == 0 && !leave)
			{
			  state[j++,k--] = 2;
			  while (len-- > 1)
			    state[j++,k--] = 3;
			  /* Since code and data is "never" mixed in 32 bit
			     code we don't need this.  */
#if 0
			  leave = (strncmp (inst, "jmp", 3) == 0
				   || strncmp (inst, "ret", 3) == 0
				   || strncmp (inst, "iret", 4) == 0);
#endif
			  if (!leave)
			    inst = unassemble_proper (j, &len);
			}
		    }
		}
	      else
		{
		  state[done] = 2;
		  next[done] = j + 1;
		}
	    }
	  j = 1;
	  k = count;
	  while (k > 0 && j <= done && state[j] >= 2)
	    if (state[j++] == 2)
	      k--;
	}
      while (k > 0 && done + 2 * count <= instcount);
      free (state);
      if (k == 0)
	return origin - j + 1;
      else
	{
	  i = origin;
	  k = 0;
	  while (count > 0 && k <= done)
	    {
	      leave = 0;
	      j = MAXINSTLEN;
	      while (!leave && j > 1)
		if (j + k <= done && next[j + k] == i)
		  leave = 1;
		else
		  j--;
	      if (!leave)
		i--, k++;
	      else
		i -= j, k += j;
	      count--;
	    }      
          free (next);
	  return i;
	}
    }
}
/* ---------------------------------------------------------------------- */
static void
code_pane_goto (word32 v)
{
  int i = 0;

  if (v >= code_dump_origin && v <= code_dump_last)
    {
      while (code_pane_pos[i] < v)
	i++;
      if (code_pane_pos [i] == code_pane_pos[i+1])
	i++;
    }
  else
    {
      code_dump_origin = v;
      if (valid_instaddr (v) && unassemble_source (v) != NULL)
	i++;
    }
  if (pane == 0)
    pane_pos = i;
  else
    pane_positions[0] = i;
}
/* ---------------------------------------------------------------------- */
inline static void
go (int bp)
{
  if (bp) activate_breakpoints ();
  is_in_running_mode = 1;
  run_child ();
  is_in_running_mode = 0;
  if (bp) deactivate_breakpoints ();
}
/* ---------------------------------------------------------------------- */
static char condition[80];

static void
step (int kind)
{
  int i, b, len, terminated, int03hit, refmem;
  char *inst = 0;

  switch (kind)
    {
    case 0:
      inst = unassemble_proper (a_tss.tss_eip, &len);
      if (strcmp (inst, "popf") == 0 || strcmp (inst, "pushf") == 0)
	kind = 1;  /* Push the right value of eflags (no trace flag).  */
      break;
    case 1:
      if (first_step)
	kind = 3;
      else
	{
	  inst = unassemble_proper (a_tss.tss_eip, &len);
	  if (strncmp (inst, "loop", 4)
	      && strncmp (inst, "call", 4)
	      && strncmp (inst, "int", 3) )
	    kind = 0;
	  break;
	}
    }
breakpoint_loop:
  switch (kind)
    {
    case 0: /* "Trace" */
      a_tss.tss_eflags |= 0x0100;
      edi.dr[7] = 0;
      refmem = strchr (inst, '[') != 0;
      if (refmem)
	/* Assume that all access to code and stack segments are safe.
	   This should hold unless you do something extra-ordinary dirty.  */
	if (((a_tss.tss_ds == a_tss.tss_cs) || (a_tss.tss_ds == a_tss.tss_ss))
	    &&
	    ((a_tss.tss_es == a_tss.tss_cs) || (a_tss.tss_es == a_tss.tss_ss))
	    &&
	    ((a_tss.tss_fs == a_tss.tss_cs) || (a_tss.tss_fs == a_tss.tss_ss)
	     || (strstr (inst, "fs:") == 0))
	    &&
	    ((a_tss.tss_gs == a_tss.tss_cs) || (a_tss.tss_gs == a_tss.tss_ss)
	     || (strstr (inst, "gs:") == 0)))
	  refmem = 0;
      if (refmem)
	user_screen ();
      go (0);
      a_tss.tss_eflags &= ~0x0100;
      break;
    case 1: /* "Step Over" */
      {
	int has_break_point = 0;
	word32 old_addr = 0;
	int code_type = 0, count = 0;

	b = get_breakpoint (BP_Code, 0, a_tss.tss_eip);
	if (b != -1)
	{
	  old_addr  = breakpoint_table[b].addr;
	  code_type = breakpoint_table[b].code_type;
	  count     = breakpoint_table[b].count;
	  strcpy(condition, breakpoint_table[b].condition);
	  reset_breakpoint (b);
	  has_break_point = 1;
	}

	user_screen ();
	b = set_breakpoint (BP_Code, 0, a_tss.tss_eip + len);
	go (1);
	reset_breakpoint (b);

	if (has_break_point)
	{
	  b = set_breakpoint (BP_Code, 0, old_addr);
	  breakpoint_table[b].code_type = code_type;
	  breakpoint_table[b].count     = count;
	  strcpy(breakpoint_table[b].condition, condition);
	}

	break;
      }
    case 2: /* "Run" */
      {
	word32 old_addr;
	int old_code_type, old_count;

	b = get_breakpoint (BP_Code, 0, a_tss.tss_eip);
	if (b != -1)
	{
	  inst = unassemble_proper (a_tss.tss_eip, &len);
	  if (strcmp (inst, "popf") == 0 || strcmp (inst, "pushf") == 0)
	  {
	    old_addr      = breakpoint_table[b].addr;
	    old_code_type = breakpoint_table[b].code_type;
	    old_count     = breakpoint_table[b].count;
	    strcpy(condition, breakpoint_table[b].condition);
	    reset_breakpoint (b);

	    b = set_breakpoint (BP_Code, 0, a_tss.tss_eip + len);
	    go (1);
	    reset_breakpoint (b);

	    b = set_breakpoint (BP_Code, 0, old_addr);
	    breakpoint_table[b].code_type = old_code_type;
	    breakpoint_table[b].count     = old_count;
	    strcpy(breakpoint_table[b].condition, condition);
	  } else
	  {
	    word32 old_dr7;

	    old_dr7 = edi.dr[7];
	    a_tss.tss_eflags |= 0x0100;
	    edi.dr[7] = 0;
	    go (0);
	    a_tss.tss_eflags &= ~0x0100;
	    edi.dr[7] = old_dr7;
	  }
	}
	user_screen ();
	go (1);
	break;
      }
    case 3: /* "Run to `_main'" */
      b = set_breakpoint (BP_Code, 0, main_entry);
      user_screen ();
      go (1);
      reset_breakpoint (b);
      break;
    }

  a_tss.tss_eflags &= ~0x0100;  /* Clear TF, in case it's set by Ctrl-C */
				/* (in remote debugging.)               */
  first_step = 0;
  i = a_tss.tss_irqn;
  terminated = (i == 0x21) && (a_tss.tss_eax & 0xff00) == 0x4c00;
  int03hit = (i == 0x03) && get_breakpoint (BP_Code, 0, a_tss.tss_eip) != -1;
  if (terminated)
    a_tss.tss_eip -= 2;  /* point back to Int 21h */
  else if (int03hit)
    a_tss.tss_eip--;     /* point back to Int 3 */
  code_pane_goto (a_tss.tss_eip);
  debug_screen ();
  redraw (0);
  if (terminated)
    message (CL_Msg, "Program terminated normally, exit code is %d",
	    (word8)a_tss.tss_eax);
  else if (i == 1 || int03hit)
    {
      int no;

      no = -1;
      for (b = 0; no == -1 && b <= 3; b++)
	if ((edi.dr[6] & (1 << b)) && (edi.dr[7] & (3 << (b * 2))))
	  no = b;

      if (no != -1)
	no = get_breakpoint ((edi.dr[7] >> (16 + 4 * no)) & 3,
			     ((edi.dr[7] >> (18 + 4 * no)) & 3) + 1,
			     edi.dr[no] - edi.app_base);
      else if (int03hit)
	no = get_breakpoint (BP_Code, 0, a_tss.tss_eip);

      if (no == -1)
	if (i == 1)
	  no = get_breakpoint(BP_Code, 0, a_tss.tss_eip);

      if ((no != -1) && (kind == 2))
	switch (breakpoint_table[no].type)
	  {
	  case BP_Code:
	    switch (breakpoint_table[no].code_type)
	    {
	      case 1:                      /* Count down */
		if (breakpoint_table[no].count)
		{
		  breakpoint_table[no].count--;
		  goto breakpoint_loop;
		} else
		  message (CL_Info, "Count down breakpoint hit");
		break;
	      case 2:                      /* Conditional */
		{
		  int sep_which, satisfy, size;
		  word32 left_value, right_value;
		  char *left_str, *right_str;

		  seperate_expr (breakpoint_table[no].condition,
				 &size, &sep_which);
		  left_str  = strdup(l_str);
		  right_str = strdup(r_str);

		  left_value  = parse_expression(size, left_str , &satisfy);
		  right_value = parse_expression(size, right_str, &satisfy);
                  free (left_str);
                  free (right_str);
		  if (sep_which != -1)
		    switch (sepers[sep_which].name)
		    {
		      case equal:
		      case equal2:
			satisfy = (left_value == right_value);
			break;
		      case great_equal:
			satisfy = (left_value >= right_value);
			break;
		      case less_equal:
			satisfy = (left_value <= right_value);
			break;
		      case not_equal:
			satisfy = (left_value != right_value);
			break;
		      case greater:
			satisfy = (left_value > right_value);
			break;
		      case less:
			satisfy = (left_value < right_value);
			break;
		    }
		  else satisfy = left_value;

		  if (satisfy == 0)
		    goto breakpoint_loop;

		  message (CL_Info, "Conditional breakpoint hit");
		  break;
		}
	      case 0:
		message (CL_Info, "Code breakpoint hit");
		break;
	    }
	    break;
	  case BP_Write:
	    message (CL_Msg, "Data write breakpoint at 0x%08lx triggered by previous instruction",
		     breakpoint_table[no].addr);
	    break;
	  case BP_Read:
	    message (CL_Msg, "Data read/write breakpoint at 0x%08lx triggered by previous instruction",
		     breakpoint_table[no].addr);
	    break;
	  }
    }
  else if (i == 0x79)
    message (CL_Info, "Keyboard interrupt");
  else if (i == 0x75)
    {
      char *reason;

      save_npx ();
      if ((npx.status & 0x0241) == 0x0241)
	reason = "stack overflow";
      else if ((npx.status & 0x0241) == 0x0041)
	reason = "stack underflow";
      else if (npx.status & 1)
	reason = "invalid operation";
      else if (npx.status & 2)
	reason = "denormal operand";
      else if (npx.status & 4)
	reason = "divide by zero";
      else if (npx.status & 8)
	reason = "overflow";
      else if (npx.status & 16)
	reason = "underflow";
      else if (npx.status & 32)
	reason = "loss of precision";
      else
	reason = "?";
      message (CL_Error,
	       "Numeric Exception (%s) at eip=0x%08lx", reason, npx.eip);
      load_npx ();
    }
  else if (i == 8 || (i >= 10 && i <= 14))
    message (CL_Error, "Exception %d (0x%02x) occurred, error code=%#lx",
	     i, i, a_tss.tss_error);
  else
    message (CL_Error, "Exception %d (0x%02x) occurred", i, i);
}
/* ---------------------------------------------------------------------- */
static void
redraw (int first)
{
  char *buf = alloca (cols);

  debug_screen ();
  screen_attr = ScreenAttrib = screen_attr_normal;

  if (first == -1)
  {
    screen_attr = screen_attr_resize;
    clear_screen();
  }

/*  if (first) */
    {
      frame (0, 1, cols - 1, rows - 1);
      frame (main_x2, 1, cols - 1, toplines + 1);
      frame (main_x1, 1, main_x2, toplines + 1);
      frame (main_x3, toplines + 1, cols - 1, rows - 1);
      frame (0, toplines + 1, main_x3, rows - 1);

      if (!using_com_port_num)
      {
	put (main_x2, 1, "");
	put (main_x2, toplines + 1, "");
	put (0, toplines + 1, "");
	put (cols - 1, toplines + 1, "");
	put (main_x1, 1, "");
	put (main_x1, toplines + 1, "");
	put (main_x3, toplines + 1, "");
	put (main_x3, rows - 1, "");
      }
    }

  if (first == -1)           /* For resizing */
  {
    put_screen (debug_screen_save);
    return;
  }

  {
    int x1, x2, y1, y2;

    screen_attr = screen_attr_ffocus;
    switch (pane)    /* Set focused pane */
    {
      case 1:        /* Top middle */
	x1 = main_x1;
	x2 = main_x2;
	y1 = 1;
	y2 = toplines + 1;
	break;
      case 2:        /* Top right */
	x1 = main_x2;
	x2 = cols - 1;
	y1 = 1;
	y2 = toplines + 1;
	break;
      case 3:        /* Bottom right */
	x1 = main_x3;
	x2 = cols - 1;
	y1 = toplines + 1;
	y2 = rows - 1;
	break;
      case 4:        /* Bottom left */
	x1 = 0;
	x2 = main_x3;
	y1 = toplines + 1;
	y2 = rows - 1;
	break;
      default:       /* Top left */
	x1 = 0;
	x2 = main_x1;
	y1 = 1;
	y2 = toplines + 1;
	break;
    }
    double_frame (x1,y1,x2,y2);
    screen_attr = screen_attr_normal;
  }

  {
    /* Show register status */
    int reg, x = main_x1 + 1, y = 2, width = main_x2 - main_x1 - 1;

    for (reg = 0; regs_type[reg]; reg++)
      {
	if (regs_type[reg] == 's')
	  sprintf (buf, " %s %04x",
		   regs_names[reg],
		   *(unsigned short *)(regs_addr[reg]));
	else
	  sprintf (buf, "%s %08lx",
		   regs_names[reg],
		   *(regs_addr[reg]));
	putl (x, y++, width, buf);
	if (y > toplines)
	  break;
      }
    while (y <= toplines)
      putl (x, y++, width, "");
  }

  {
    /* Show flags status */
    int f, x = main_x2 + 1, y = 2, width = cols - main_x2 - 2;
    static char flags[] = "c?p?a?zn?ido????";

    for (f = 0; flags[f]; f++)
    {
      if (flags[f] != '?')
	{
	  sprintf (buf, "%c=%d", flags[f], (int)((a_tss.tss_eflags >> f) & 1));
	  putl (x, y++, width, buf);
	}
      if (y > toplines)
	break;
    }
    while (y <= toplines)
      putl (x, y++, width, "");
  }

  {
    /* Show breakpoints */
    int b, x = main_x3 + 1, y = toplines + 2, width = cols - main_x3 - 2;
    char *name;
    int32 delta;

    for (b = breakpoint_origin;
	 b < breakpoint_origin + bottomlines / 2 && b < breakpoint_count;
	 b++)
      {
	switch (breakpoint_table[b].type)
	  {
	  case BP_Code:
	    sprintf (buf,(breakpoint_table[b].code_type == 0
			  ? "Execute   :" :
			   (breakpoint_table[b].code_type == 1
			    ? "Count-down:" : "Conditonal:")));
	    break;
	  case BP_Write:
	    sprintf (buf, "%-20s, %d","Data write:",
		     breakpoint_table[b].length);
	    break;
	  case BP_Read:
	    sprintf (buf, "%-20s, %d","Data read :",
		     breakpoint_table[b].length);
	    break;
	  }
	putl (x, y, width, buf);
	screen_attr = screen_attr_addr;
	sprintf(buf, "%08lx", breakpoint_table[b].addr);
	if (width > 12)
	  putl(x + 12, y, width - 12 > 8 ? 8 : width - 12, buf);
	y++;
	screen_attr = screen_attr_normal;

	name = syms_val2name (breakpoint_table[b].addr, &delta);
	if (name[0] != '0')
	  if (delta && strlen (name) < width)
	    sprintf (buf, " %s+%#lx", name, delta);
	  else
	    sprintf (buf, " %-*s", width, name);
	else
	  buf[0] = '\0';
	putl (x, y++, width , buf);
      }
    if (breakpoint_count == 0)
      putl (x, y++, width, "No breakpoints");
    while (y < rows - 1)
      putl (x, y++, width, "");
  }

  {
    /* Show data dump */
    word32 p = data_dump_origin;
    int b, x, y = toplines + 2, ok, width = main_x3 - 1;
    unsigned char data[8], xpos[8];

    for (x = 0; x < 8; x++)
      xpos[x] = 1
	+ ((data_dump_size == 1) ? (x * 3) :
	   ((data_dump_size == 2) ? ((x / 2) * 5 + (~x & 1) * 2) :
	    ((x / 4) * 9 + (~x & 3) * 2)));

    while (y < rows - 1)
      {
	if ((ok = valid_addr (p, 8)))
	  read_child (p, data, 8);

	screen_attr = screen_attr_addr;
	sprintf (buf, "%08lx:",p);
	putl (1, y, 9, buf);
	sprintf (buf, "%70s", " ");
	for (x = 0; x < 8; x++)
	  if (ok || valid_addr (p + x, 1))
	    {
	      if (!ok) read_child (p + x, data + x, 1);
	      buf[xpos[x]] = hexchars[data[x] >> 4];
	      buf[xpos[x] + 1] = hexchars[data[x] & 0xf];
	      buf[28 + x] = (data[x] ? data[x] : '.');
	    }
	  else
	    buf[xpos[x]] = buf[xpos[x + 1]] = buf[28 + x] = '?';
	screen_attr = screen_attr_normal;
	putl (10, y, width - 9, buf);

	screen_attr = screen_attr_break;
	for (x = 0; x < 8; x++)
	  for (b = 0; b < breakpoint_count; b++)
	    if (breakpoint_table[b].type != BP_Code
		&& p + x >= breakpoint_table[b].addr
		&& p + x < (breakpoint_table[b].addr
			    + breakpoint_table[b].length))
	      {
		highlight (xpos[x] + 10, y, 2);
		highlight (38 + x, y, 1);
	      }
	screen_attr = screen_attr_normal;
	p += 8;
	y++;
      }
    data_dump_last = p - 1;
  }

  if (code_pane_active)
  {
    /* Show code dump */
    word32 p = code_dump_origin, width = main_x1 - 1;
    int y = 2, len, source = 0;
    char *txt;

    while (y <= toplines)
      {
	source = !source;
	code_pane_pos[y - 2] = p;
	if (source)
	  {
	    txt = unassemble_source (p);
	    if (txt)
	      {
		screen_attr = screen_attr_source;
		putl (1, y++, width, txt);
         /*       free (txt);     */
	      }
	  }
	else
	  {
	    int temp_len;

	    if (valid_instaddr (p))
	      txt = unassemble_proper (p, &len);
	    else
	      txt = "?", len = 1;
	    sprintf (buf, "%08lx%c",
		     p,
		     p == a_tss.tss_eip ? '>' : ' ');
	    screen_attr =
	      (get_breakpoint (BP_Code, 0, p) == -1
	       ? screen_attr_addr : screen_attr_break);
	    temp_len = strlen(buf);
	    putl (1, y++, (width > temp_len ? temp_len : width), buf);
	    if (width > temp_len)
	    {
	      screen_attr =
		(get_breakpoint (BP_Code, 0, p) == -1
		 ? screen_attr_asm : screen_attr_break);
	      putl(temp_len + 1,y - 1,(width - temp_len),txt);
	    }
	    p += len;
	  }
      }
    code_pane_pos[y - 2] = p + 1;
    code_dump_last = p - 1;
    screen_attr = screen_attr_normal;
  }

  if (npx_pane_active)
  {
    /* Show npx dump */
    int i, y = 2, width = main_x1 - 1;
    static char *rtype[] = { "Near", "-Inf", "+Inf", "Zero" };

    save_npx ();
    sprintf (buf,
	     "Control: %04lx PR=%s UN=%s OV=%s ZD=%s DN=%s IV=%s Rnd=%s",
	     npx.control & 0xffff,
	     (npx.control & (1 << 5)) ? "N" : "Y",
	     (npx.control & (1 << 4)) ? "N" : "Y",
	     (npx.control & (1 << 3)) ? "N" : "Y",
	     (npx.control & (1 << 2)) ? "N" : "Y",
	     (npx.control & (1 << 1)) ? "N" : "Y",
	     (npx.control & (1 << 0)) ? "N" : "Y",
	     rtype[(npx.control >> 10) & 3]);
    putl (1, y++, width, buf);
    sprintf (buf,
	     "Status:  %04lx PR=%s UN=%s OV=%s ZD=%s DN=%s IV=%s ST=%s",
	     npx.status & 0xffff,
	     (npx.status & (1 << 5)) ? "Y" : "N",
	     (npx.status & (1 << 4)) ? "Y" : "N",
	     (npx.status & (1 << 3)) ? "Y" : "N",
	     (npx.status & (1 << 2)) ? "Y" : "N",
	     (npx.status & (1 << 1)) ? "Y" : "N",
	     (npx.status & (1 << 0)) ? "Y" : "N",
	     (npx.status & (1 << 6)) ? "Y" : "N");
    if (y <= toplines)
      putl (1, y++, width, buf);
    sprintf (buf, "%19sC3=%d C2=%d C1=%d C0=%d",
	     "",
	     (npx.status & (1 << 14)) != 0,
	     (npx.status & (1 << 10)) != 0,
	     (npx.status & (1 <<  9)) != 0,
	     (npx.status & (1 <<  8)) != 0);
    if (y <= toplines)
      putl (1, y++, width, buf);

    for (i = 0; i < 8; i++)
      {
	/* We assume that `long double' is the same type as npx.reg[i].
	   It would be sensible to check that the sizes match, but they
	   don't!  For alignment reasons, `sizeof (long double)' is 12.  */
	long double d;
	int tag;
	int tos = (npx.status >> 11) & 7;
	int exp = (int)npx.reg[i].exponent - 16382;
	char *dstr = alloca (30);

	dstr[0] = (npx.reg[i].sign) ? '-' : '+';
	dstr[1] = '\0';
	tag = (npx.tag >> (((i + tos) & 7) * 2)) & 3;
	switch (tag)
	  {
	  case 0:
	    if (abs (exp) < 1000)
	      {
		d = *((long double*)(npx.reg + i));
		/* sprintf does not (djgpp 1.11m3) handle long doubles.  */
		sprintf(dstr,"%+.16g", (double) d);
	      }
	    else
	      sprintf (dstr, "Valid, %s, and %s",
		       npx.reg[i].sign ? "negative" : "positive",
		       exp > 0 ? "huge" : "tiny");
	    break;
	  case 1:
	    strcat (dstr, "Zero");
	    break;
	  case 2:
	    if (npx.reg[i].exponent == 0x7fff)
	      if (npx.reg[i].sig3 == 0x8000
		  && npx.reg[i].sig2 == 0x0000
		  && npx.reg[i].sig1 == 0x0000
		  && npx.reg[i].sig0 == 0x0000)
		strcat (dstr, "Inf");
	      else
		strcat (dstr, "NaN");
	    else
	      dstr = "Special";
	    break;
	  case 3:
	    dstr = "Empty";
	    break;
	  }
	sprintf (buf, "st(%d): %d %04x %04x%04x%04x%04x %s",
		 i,
		 npx.reg[i].sign, npx.reg[i].exponent,
		 npx.reg[i].sig3, npx.reg[i].sig2,
		 npx.reg[i].sig1, npx.reg[i].sig0,
		 dstr);
        free (dstr);
	putl (1, y++, width, buf);
	if (y > toplines)
	  break;
      }
    while (y <= toplines)
      putl (1, y++, width, "");
    load_npx ();
  }

  if (stack_pane_active)
  {
    /* Show stack dump */
    int line, y = 2, no = -stack_dump_origin, width = main_x1 - 1;
    unsigned char eipcode[4];
    word32 delta, v = a_tss.tss_ebp;
    word32 vaddr = a_tss.tss_eip;

    stack_dump_more = 0;
    while (valid_instaddr (vaddr))
      {
	if (no++ >= 0)
	  if (y >= toplines)
	    {
	      stack_dump_more = 1;
	      break;
	    }
	  else
	    {
	      char *symaddr, *sourceaddr, *sourcefile;
	      int len;

	      symaddr = syms_val2name (vaddr, &delta);
	      sourcefile = syms_val2line (vaddr, &line, 0);
	      if (sourcefile)
		sprintf (sourceaddr = alloca (cols + strlen (sourcefile)),
			 ", line %d in file %s", line, sourcefile);
	      else if (delta)
		sprintf (sourceaddr = alloca (cols), "%+ld", (long) delta);
	      else
		sourceaddr = "";
	      len = 15 + strlen (symaddr) + strlen (sourceaddr);
	      if (len >= cols)
	      {
		free (buf);
		buf = alloca (len);
	      }
	      sprintf (buf, "%08lx: %s%s", vaddr, symaddr, sourceaddr);
	      stack_dump_last = (y - 2) / 2;
	      stack_dump_pos[stack_dump_last] = vaddr;
	      putl (1, y++, width, buf);
	      sprintf (buf, "   Stack frame: %08lx", v);
	      putl (1, y++, width, buf);
	      free (symaddr);
	      free (sourceaddr);
	      free (sourcefile);
	    }

	read_child (vaddr, eipcode, 3);
	/* We look directly at the bit pattern instead of using disassembly;
	   the bit patterns are those generated by gcc.  In general we
	   cannot expect to be able to follow a non-gcc stack anyway.  */
	if ((eipcode[0] == 0x55 && eipcode[1] == 0x89 && eipcode[2] == 0xe5)
	    || eipcode[0] == 0xc3)
	  {
	    /* In this case where we are looking at `Push Ebp//Mov Ebp,Esp'
	       or `Ret', only the return address is on the stack; I believe
	       this only to happen in the innermost activation record.  */
	    if (valid_addr (a_tss.tss_esp, 4))
	      read_child (a_tss.tss_esp, &vaddr, 4);
	    else
	      break;
	  }
	else
	  {
	    if (eipcode[0] == 0x89 && eipcode[1] == 0xe5)
	      /* When looking at `Mov Esp,Ebp' the next stack frame is on
		 the stack and pointed to by the stack pointer.  We are
		 actually in the same situation after `Mov Esp,Ebp' which
		 is generated by gcc with the -m486 option.  This case is,
		 however, handled perfectly by using the base pointer.  */
	      v = a_tss.tss_esp;
	    if (v >= a_tss.tss_esp && v < 0x80000000L && valid_addr (v, 8))
	      {
		read_child (v + 4, &vaddr, 4);
		read_child (v, &v, 4);
	      }
	    else
	      break;
	  }
      }
    while (y <= toplines)
      putl (1, y++, width, "");
  }

  if (info_pane_active)
  {
    /* Show info */
    int y = 2, width = main_x1 - 1;
    long ul;
    char *s = alloca (20);
    _go32_dpmi_meminfo info;
    _go32_dpmi_registers regs;

    sprintf (buf, "Debugger version ............: %d.%02d %s",
	     FS_MAJOR_VER, FS_MINOR_VER,
	     FS_MAJOR_VER == 0 ? "beta" : "");
    putl (1, y++, width, buf);
    _go32_dpmi_get_free_memory_information (&info);
    switch (_go32_info_block.run_mode)
      {
      case _GO32_RUN_MODE_RAW:
	sprintf (s, "Raw");
	break;
      case _GO32_RUN_MODE_XMS:
	sprintf (s, "Xms");
	break;
      case _GO32_RUN_MODE_VCPI:
	sprintf (s, "Vcpi");
	break;
      case _GO32_RUN_MODE_DPMI:
	sprintf (s, "Dpmi %d.%02x",
		 _go32_info_block.run_mode_info >> 8,
		 _go32_info_block.run_mode_info & 0xff);
	break;
      default:
	sprintf (s, "Unknown");
      }
    sprintf (buf, "Running mode ................: %s", s);
    free (s);
    if (y <= toplines)
      putl (1, y++, width, buf);
    if (y <= toplines)
      putl (1, y++, width, "");

    ul = info.total_physical_pages;
    if (ul > 0)
      {
	ul <<= 12;
	sprintf (buf, "Total physical memory .......: %lu KB", ul >> 10);
	if (y <= toplines)
	  putl (1, y++, width, buf);
      }
    ul = info.available_physical_pages
      ? info.available_physical_pages << 12
	: info.available_memory;
    sprintf (buf, "Remaining physical memory ...: %lu KB", ul >> 10);
    if (y <= toplines) putl (1, y++, width, buf);
    ul = info.available_memory;
    sprintf (buf, "Remaining virtual memory ....: %lu KB", ul >> 10);
    if (y <= toplines)
      putl (1, y++, width, buf);
    ul = info.free_linear_space << 12;
    if (ul >= 0)
      {
	sprintf (buf, "Free linear space ...........: %ld KB", ul >> 10);
	if (y <= toplines)
	  putl (1, y++, width, buf);
      }
    /* Remember that Dos memory is only made available of direct request;
       using 0xffff does not count as a request.  */
    regs.h.ah = 0x48;
    regs.x.bx = 0xffff;
    regs.x.ss = regs.x.sp = 0;                        
/* ----------------------------------------------------------
   -  The call _go32_dpmi_simulate_int () causes the program 
   -  to hang under XMS mode.
   ---------------------------------------------------------- */
    if (_go32_info_block.run_mode != _GO32_RUN_MODE_XMS)
      _go32_dpmi_simulate_int (0x21, &regs);
    ul = regs.x.bx << 4;
    sprintf (buf, "Free dos memory .............: %ld %s",
	     ul > 8192 ? ul >> 10 : ul,
	     ul > 8192 ? "KB"     : "Bytes");
    if (y <= toplines)
      putl (1, y++, width, buf);
    if (y <= toplines)
      putl (1, y++, width, "");
    sprintf (buf, "Ctrl-C checking .............: %s",
	     getcbrk () ? "On" : "Off");
    if (y <= toplines)
      putl (1, y++, width, buf);
    if (y <= toplines)
      putl (1, y++, width, "");

    sprintf (buf, "Program text ................: %08lx - %08lx",
	     areas[A_text].first_addr, areas[A_text].last_addr);
    if (y <= toplines)
      putl (1, y++, width, buf);
    sprintf (buf, "Program data ................: %08lx - %08lx",
	     areas[A_data].first_addr, areas[A_data].last_addr);
    if (y <= toplines)
      putl (1, y++, width, buf);
    sprintf (buf, "Program bss .................: %08lx - %08lx",
	     areas[A_bss].first_addr, areas[A_bss].last_addr);
    if (y <= toplines)
      putl (1, y++, width, buf);
    sprintf (buf, "Program stack ...............: %08lx - %08lx",
	     areas[A_stack].first_addr, areas[A_stack].last_addr);
    if (y <= toplines)
      putl (1, y++, width, buf);

/* ----------------- Testing -com_num switch ------------- */
    sprintf (buf, "Using com port number : %d", using_com_port_num);
    if (y <= toplines)
      putl (1, y++, width, buf);
/* ----------------- Testing -com_num switch ------------- */

    while (y <= toplines)
      putl (1, y++, width, "");
  }

  if (whereis_pane_active)
  {
    /* Show result of last whereis command */
    int y = 2, width = main_x1 - 1, i = whereis_origin;

    while (y <= toplines)
      if (i < whereis_text_count)
	putl (1, y++, width, whereis_pane_text[i++]);
      else
	putl (1, y++, width, "");
  }

  if (module_pane_active)
  {
    int y = 2, width = main_x1 - 1, i = module_origin;

    while (y <= toplines)
      if (i < module_text_count)
	putl (1, y++, width, module_pane_text[i++]);
      else
	putl (1, y++, width, "");
  }

  if (help_pane_active)
  {
    int y = 2, width = main_x1 - 1, i = help_origin;

    while (y <= toplines)
      if (i < help_text_count)
      {
	screen_attr =
	  (*help_pane_text[i++] != ' '
	    ? screen_attr_help : screen_attr_normal);
	putl (1, y++, width, help_pane_text[i-1]);
      }
      else
	putl (1, y++, width, "");
  }

  {
    /* Highlight focus */
    screen_attr = screen_attr_focus;

    switch (pane)
      {
      case 6: /* Stack */
	if (pane_pos > stack_dump_last)
	  pane_pos = stack_dump_last;
	highlight (1, pane_pos * 2 + 2, main_x1 - 1);
	break;
      case 0: /* Code */
      case 5: /* NPX */
	highlight (1, pane_pos + 2, main_x1 - 1);
	break;
      case 1: /* Registers */
	highlight (main_x1 + 1, pane_pos + 2, main_x2 - main_x1 - 1);
	break;
      case 2: /* Flags */
	highlight (main_x2 + 1, pane_pos + 2, cols - main_x2 - 2);
	break;
      case 3: /* Breakpoints */
	if (breakpoint_origin >= breakpoint_count)
	  breakpoint_origin = breakpoint_count ? breakpoint_count - 1 : 0;
	if (breakpoint_origin + pane_pos >= breakpoint_count)
	  pane_pos = breakpoint_count - breakpoint_origin;
	highlight (main_x3 + 1, toplines + 2 + 2 * pane_pos,
		   cols - main_x3 - 2);
	highlight (main_x3 + 1, toplines + 3 + 2 * pane_pos,
		   cols - main_x3 - 2);
	break;
      case 4: /* Data */
	{
	  int x = pane_pos & 7, y = pane_pos >> 3;

	  if (38 + x + data_dump_size <= main_x3)
	     highlight (38 + x, toplines + 2 + y, data_dump_size);
	  switch (data_dump_size)
	    {
	    case 1:
	      if (13 + 3 * x <= main_x3)
		highlight (11 + 3 * x, toplines + 2 + y, 2);
	      break;
	    case 2:
	      if (15 + 5 * (x >> 1) <= main_x3)
		highlight (11 + 5 * (x >> 1), toplines + 2 + y, 4);
	      break;
	    case 4:
	      if (19 + 9 * (x >> 2) <= main_x3)
		highlight (11 + 9 * (x >> 2), toplines + 2 + y, 8);
	      break;
	    }
	}
	break;
      case 8: /* Whereis */
	if (whereis_text_count)
	  highlight (1, pane_pos + 2, main_x1 - 1);
	break;
      case 9: /* Module */
	if (module_text_count)
	  highlight (1, pane_pos + 2, main_x1 - 1);
	break;
      case 10: /* Help */
	if (help_text_count)
	  highlight (1, pane_pos + 2, main_x1 - 1);
	break;
      }
  }
  free (buf);
  put_screen (debug_screen_save);
  { int old_menu_pos, temp_bar_sel;

    old_menu_pos = Main_Menu[0].type;
    Main_Menu[0].type = 0;
    display_bar_menu (Main_Menu, &temp_bar_sel);
    Main_Menu[0].type = old_menu_pos;
  }
}
/* ---------------------------------------------------------------------- */
static void
init_module()
{
  char *s;

  s = malloc(25);
  s = "Modules listing:";
  module_pane_text = realloc (module_pane_text,
			  (++module_text_count) * sizeof (char *));
  module_pane_text[module_text_count - 1] = strdup (s);
  free(s);
  s = get_module(module_text_count - 1);
  while (!undefined_symbol)
  {
    module_pane_text = realloc (module_pane_text,
			  (++module_text_count) * sizeof (char *));
    module_pane_text[module_text_count - 1] = strdup (s);
    free(s);
    s = get_module(module_text_count - 1);
  }
  undefined_symbol = 0;
}
/* ---------------------------------------------------------------------- */
static void
put_help_line(const char *str)
{
  help_pane_text = realloc (help_pane_text,
			   (++help_text_count) * sizeof (char *));
  help_pane_text[help_text_count - 1] = strdup (str);

}
/* ---------------------------------------------------------------------- */
static void
init_help()
{
  put_help_line("Global commands");
  put_help_line("---------------");
  put_help_line(" Up/Left      : Move up one line/byte/word/dword.");
  put_help_line(" Down/Right   : Move down one line/byte/word/dword.");
  put_help_line(" PageUp/Down  : Move up/down one page.");
  put_help_line(" Ctrl-Y       : Move up one page.");
  put_help_line(" Ctrl-V       : Move down one page.");
  put_help_line(" Home/End     : Move to the beginning/end of pane.");
  put_help_line(" ");
  put_help_line(" Ctrl-C       : Menu.");
  put_help_line(" Ctrl-R       : Run.");
  put_help_line(" Ctrl-F       : Trace into function.");
  put_help_line(" Ctrl-O       : Step over one machine instruction.");
  put_help_line(" Ctrl-A       : Animate - run program at specified speed.");
  put_help_line(" Ctrl-Z       : Refresh screen.");
  put_help_line(" Ctrl-X       : Local menu (If the current pane has one.)");
  put_help_line(" Tab          : Next pane.");
  put_help_line(" Shift-Tab    : Previous pane.");
  put_help_line(" ");
  put_help_line("Help pane");
  put_help_line("---------");
  put_help_line(" Space        : Return to code pane.");
  put_help_line(" Enter/Escape : Return to code pane.");
  put_help_line(" ");
  put_help_line("Code pane");
  put_help_line("---------");
  put_help_line(" Ctrl-B       : Toggle breakpoint.");
  put_help_line(" Ctrl-T       : Trace to current address.");
  put_help_line("                Same as setting a normal breakpoint");
  put_help_line("                then pressing <run>.");
  put_help_line(" Ctrl-G       : Disassemble and display memory at");
  put_help_line("                the specified address");
  put_help_line(" Ctrl-E       : Disassemble and display memory at");
  put_help_line("                address specified by EIP.");
  put_help_line(" Ctrl-I       : Set EIP to current address.");
  put_help_line(" Ctrl-H       : Modify performance analyzer settings.");
  put_help_line(" Ctrl-P       : Start performance analyzer and run.");
  put_help_line("                (Only works if the timer has been set");
  put_help_line("                 using Ctrl-E.)");
  put_help_line(" Ctrl-K       : Start disassembling one byte earlier.");
  put_help_line(" Ctrl-L       : Start disassembling one byte later.");
  put_help_line(" ");
  put_help_line("Data pane");
  put_help_line("---------");
  put_help_line(" Ctrl-U       : Toggle data write breakpoint.");
  put_help_line(" Ctrl-T       : Toggle data read/write breakpoint.");
  put_help_line(" Ctrl-B/W/D   : Display in Byte/Word/DWord format.");
  put_help_line(" Ctrl-G       : Display memory at specified address.");
  put_help_line(" Ctrl-E       : Display memory at address specified by EIP.");
  put_help_line(" Ctrl-H       : Display memory at address specified by ESP.");
  put_help_line(" Ctrl-N       : Display data in customized format.");
  put_help_line(" Ctrl-P       : Input/Output from/to ports, as follows:");
  put_help_line("                   Action - input or output");
  put_help_line("                   Size   - 1 = byte, 2 = word, 4 = dword.");
  put_help_line("                   Port   - The port to read/write.");
  put_help_line("                   Data   - Data to be written to the port.");
  put_help_line(" <number>     : Modify data at cursor.");
  put_help_line(" <string>     : Modify data at cursor (in byte mode only).");
  put_help_line(" Ctrl-K       : Start displaying one byte earlier.");
  put_help_line(" Ctrl-L       : Start displaying one byte later.");
  put_help_line(" ");
  put_help_line("Registers pane");
  put_help_line("--------------");
  put_help_line(" <number>     : Modify register at cursor.");
  put_help_line(" ");
  put_help_line("Flags pane");
  put_help_line("----------");
  put_help_line(" Space/+/-/t  : Toggle flag at cursor.");
  put_help_line(" s, 1         : Set flag at cursor.");
  put_help_line(" r, 0         : Clear flag at cursor.");
  put_help_line(" ");
  put_help_line("Breakpoint pane");
  put_help_line("---------------");
  put_help_line(" Enter/Ctrl-G : Display breakpoint.");
  put_help_line(" Del / Ctrl-D : Delete breakpoint.");
  put_help_line(" Ctrl-E       : Edit code breakpoint.");
  put_help_line(" ");
  put_help_line("Whereis pane");
  put_help_line("------------");
  put_help_line(" Enter        : Display memory/code at address represented");
  put_help_line("                by the currently selected symbol.");
  put_help_line(" <string>     : Look up <string> in symbol table.");
  put_help_line("                Permissible wildcards are:");
  put_help_line("                   ? - matches one character. ");
  put_help_line("                   * - matches any number of characters. ");
  put_help_line(" ");
  put_help_line("Module pane");
  put_help_line("-----------");
  put_help_line(" Enter        : Promt for a line number, and display.");
  put_help_line("                code at that location.");
  put_help_line(" ");
  put_help_line("Edit Breakpoint Window");
  put_help_line("----------------------");
  put_help_line(" Enter        : At <Ok>/<Cancel> - Accept/Discard changes.");
  put_help_line("                Otherwise, edit currently selected item.");
  put_help_line("                Editable parameters are listed below.");
  put_help_line("   Address    : The location of the breakpoint.");
  put_help_line("   Type       : The breakpoint's type.");
  put_help_line("                Permissible types are:");
  put_help_line("                   0 - Normal breakpoint, always enabled.");
  put_help_line("                   1 - Enable breakpoint after <count> hits.");
  put_help_line("                   2 - Enable breakpoint if <condition>");
  put_help_line("                       is true.");
  put_help_line("   Count      : Hits before a type 1 breakpoint activates.");
  put_help_line("   Condition  : Condition for enabling a type 2 breakpoint.");
  put_help_line("                Syntax is as follows:");
  put_help_line("                   [<size>] <expr> [<operator> <expr>]");
  put_help_line("                where");
  put_help_line("                  <size> - either 1, 2 or 4. Denotes maximum");
  put_help_line("                   data size to use while evaluating <expr>.");
  put_help_line("                  <expr> - a valid C expression which uses");
  put_help_line("                   no parentheses (is not nested) and uses");
  put_help_line("                   only global variable names, symbolic");
  put_help_line("                   constants, constant values, and register");
  put_help_line("                   names. All register names must begin with");
  put_help_line("                   a '%' (percent) character.");
  put_help_line("                  <operator> - a binary relational operator");
  put_help_line("                   such as:");
  put_help_line("                      <  <=  ==  !=  <>  =  =>  >");
  put_help_line("                Example of a condition:");
  put_help_line("                   (4)!%ecx == %eax << 2 + 4");
  put_help_line(" Escape       : Discard changes (but keep the breakpoint).");
  put_help_line(" ");
  put_help_line("Npx pane");
  put_help_line("--------");
  put_help_line(" Ctrl-E       : Empty register.");
  put_help_line(" Ctrl-N       : Negate register.");
  put_help_line(" Ctrl-Z       : Zero register.");
  put_help_line(" <number>     : Modify register at cursor.");
  put_help_line(" ");
  put_help_line("Performance Analyzer Window");
  put_help_line("---------------------------");
  put_help_line(" Enter        : At <On>/<Off>  - Toggle timer.");
  put_help_line("                At <ReInit>    - Reinitialize address range.");
  put_help_line("                At <View>      - View data (if any).");
  put_help_line("                At <Done>      - Return to code pane.");
  put_help_line(" While viewing, use Ctrl-L to log and anykey to continue.");
  put_help_line(" ");
  put_help_line("IO Redirection");
  put_help_line("--------------");
  put_help_line(" Use EXTREME CAUTION when redirecting input and output.");
  put_help_line(" Incorrect values WILL result in loss of debugger control!");
  put_help_line(" ");
  put_help_line(" Possible values are:");
  put_help_line("    con       : Redirect  to console.");
  put_help_line("    com_one   : Redirect IO to COM1.");
  put_help_line("    com_two   : Redirect IO to COM2.");
  put_help_line("    com_three : Redirect IO to COM3.");
  put_help_line("    com_four  : Redirect IO to COM4.");
}
/* ---------------------------------------------------------------------- */
static void
initialize ()
{
  int i;

  using_com_port_num = edi.com_port;
  debug_screen_p = 0;
  transfer_screen_save = 0;
  max_cols = 80;
  max_rows = 50;
  if (using_com_port_num)
  {
    cols = 80;
    rows = 24;
    com_init(using_com_port_num);                   
  }
  else
  {
    max_cols = ScreenCols ();
    max_rows = ScreenRows ();
    cols = max_cols;
    rows = max_rows;
  }
  if (cols < 80 || rows < 24)
    {
      fprintf (stderr, "\nDebugger error:\n\
There are only %d columns and %d rows\n\
in this display mode.\n\
The debugger needs at least\n\
80 columns and 24 rows.\n",
	       cols, rows);
      exit (1);
    }
  toplines = (rows / 2) + 4;
  bottomlines = rows - 3 - toplines;
  main_x1 = cols - 18;
  main_x2 = cols - 5;
  main_x3 = 46;

  my_line = malloc (4096);                          
  small_line = malloc (40);
  read_buffer = malloc (cols + 10);
  code_pane_pos = malloc ((rows + 2) * sizeof (word32));
  stack_dump_pos = malloc ((rows + 2) * sizeof (word32));

  debug_screen_save = get_screen ();
  debug_screen_save[1] = debug_screen_save[2] = 0;  /* Patch cursor pos.  */

  transfer_screen_save = get_screen ();
  transfer_screen_save[1] = transfer_screen_save[2] = 0;

  pane = 10;
  pane_pos = 0;
  for (i = 0; i < PANECOUNT; i++) pane_positions[i] = 0;
  data_dump_origin = areas[A_data].first_addr;
  data_dump_size = 1;
  code_dump_origin = a_tss.tss_eip;
  stack_dump_origin = 0;
  breakpoint_origin = 0;
  breakpoint_count = 0;
  breakpoint_table = malloc (breakpoint_count * sizeof (BP_ENTRY));
  whereis_origin = 0;
  whereis_text_count = 0;
  whereis_pane_text = malloc (whereis_text_count * sizeof (char *));

  module_origin = 0;
  module_text_count = 0;
  module_pane_text = malloc (module_text_count * sizeof (char *));
  init_module ();

  help_origin = 0;
  help_text_count = 0;
  help_pane_text = malloc (help_text_count * sizeof (char *));
  init_help();

  help_pane_active = 1;
  npx_pane_active = stack_pane_active = info_pane_active
    = whereis_pane_active = module_pane_active
    = code_pane_active = 0;

  switch (ScreenMode ())
    {
    case 2:
    case 7:
      /* Mono */
      screen_attr_normal   = (A_black << 4) + A_grey;
      screen_attr_source   = screen_attr_normal;
      screen_attr_focus    = (A_grey  << 4) + A_black;
      screen_attr_ffocus   = (A_grey  << 4) + A_black;
      screen_attr_break    = (A_black << 4) + A_white;
      screen_attr_message  = (A_grey  << 4) + A_white;
      screen_attr_error    = (A_grey  << 4) + A_black;
      screen_attr_resize   = (A_grey  << 4) + A_black;
      screen_attr_asm      = screen_attr_normal;
      screen_attr_help     = screen_attr_normal;
      screen_attr_addr     = screen_attr_normal;
      break;
    default:
      /* Colour */
      screen_attr_normal   = (A_grey  << 4) + A_blue;
      screen_attr_source   = (A_grey  << 4) + A_black;
      screen_attr_focus    = (A_blue  << 4) + A_white;
      screen_attr_ffocus   = (A_cyan  << 4) + A_white;
      screen_attr_break    = (A_red   << 4) + A_white;
      screen_attr_message  = (A_green << 4) + A_white;
      screen_attr_error    = (A_red   << 4) + A_white;
      screen_attr_resize   = (A_grey  << 4) + A_green;
      screen_attr_asm      = (A_grey  << 4) + A_brown;
      screen_attr_help     = (A_grey  << 4) + A_yellow;
      screen_attr_addr     = (A_grey  << 4) + A_bold + A_black;
    }

  init_eval_data (0, 0); /* Force to calculate addresses */

  redraw (1);
  message (CL_Info, "Sally Full Screen Debugger - Modified by Long Doan");
}
/* ---------------------------------------------------------------------- */
static void
view_eval_data ()
{
  EVAL_SEG_INFO swap_data;
  EVAL_DATA_TABLE temp_data;
  int i, j, percent_bar;
  long percent;
  int32 delta, delta2;
  word32 total_hits = 0, in_hits = 0, out_hits;
  char *save, *name, *name2, *p = alloca (256);

  if (total_hits_count == 0)
  {
    message (CL_Error,
	     "Nothing has been recorded, cannot calculate percentage");
    free (p);
    return;
  }

  for (i = 0; i < EVAL_SEGMENTS; i++)
  {
    temp_data[i].start_addr = evaluation_data[i].start_addr;
    temp_data[i].stop_addr  = evaluation_data[i].stop_addr;
    temp_data[i].hits_count = evaluation_data[i].hits_count;

    total_hits += temp_data[i].hits_count;
  }

  for (i = 0; i < EVAL_SEGMENTS - 1; i++)
    for (j = i + 1; j < EVAL_SEGMENTS; j++)
      if (temp_data[i].hits_count < temp_data[j].hits_count)
      {
	swap_data.start_addr = temp_data[i].start_addr;
	swap_data.stop_addr  = temp_data[i].stop_addr;
	swap_data.hits_count = temp_data[i].hits_count;

	temp_data[i].start_addr = temp_data[j].start_addr;
	temp_data[i].stop_addr  = temp_data[j].stop_addr;
	temp_data[i].hits_count = temp_data[j].hits_count;

	temp_data[j].start_addr = swap_data.start_addr;
	temp_data[j].stop_addr  = swap_data.stop_addr;
	temp_data[j].hits_count = swap_data.hits_count;
      }

  j = (rows / 2) - 4;

  save = debug_screen_save;
  debug_screen_save = get_screen ();

  screen_attr = screen_attr_normal;
  clear_screen ();

  screen_attr = screen_attr_ffocus;
  double_frame (0, 0, cols - 1, rows - 1);
  screen_attr = screen_attr_normal;

  sprintf (p, "Most called segments (from %08lx to %08lx)",
        evaluation_data[0].start_addr, 
        evaluation_data[EVAL_SEGMENTS - 1].stop_addr);
  putl (cols / 2 - 24, 1, 50, p);

  for (i = 0; i < j; i++)
    in_hits += temp_data[i].hits_count;

  if (in_hits == 0)
    putl (2, 3, cols-5, "Nothing is recorded inside the given addresses.");
  else
  {
    for (i = 0; i < j; i++)
    {                                                                           
      percent = (long)temp_data[i].hits_count;
      percent *= 100;

      name = syms_val2name (temp_data[i].start_addr, &delta);
      name2 = syms_val2name (temp_data[i].stop_addr, &delta2);
      if (name[0] != '0')
	if (delta)
	  sprintf (p, "(%s+%#lx)", name, delta);
	else
	  sprintf (p, "(%s)", name);
      else
	sprintf(p, "(No Symbols");
      free (name);
      name = strdup (p);
      if (name2[0] != '0')
	if (delta2)
	  sprintf (p, "(%s+%#lx)", name2, delta2);
	else
	  sprintf (p, "(%s)", name2);
      else
	sprintf(p, "No Symbols)");
      free (name2);
      name2 = strdup (p);

      sprintf (p, "%02d.%-1d%% %08lx-%08lx %s-%s",
	       percent / total_hits, ((percent % total_hits) * 10) / total_hits,
               temp_data[i].start_addr, temp_data[i].stop_addr,
	       name, name2);
      free (name);
      free (name2);                            
      percent /= 100;
      percent_bar = percent * (cols - 5) / total_hits;

      putl (2, i * 2 + 3, cols - 4, p);
      draw (2, i * 2 + 4, '-', 1, percent_bar);

    }

    out_hits = total_hits - in_hits;
    percent = out_hits * 100;

    sprintf (p, "%02d.%-1d%% : Other segments (%% of total in-range hits)", 
             percent / total_hits, ((percent % total_hits) * 10) / total_hits);
    putl (2, i * 2 + 3, cols - 4, p);           
    percent /= 100;
    percent_bar = percent * (cols - 5) / in_hits;
    draw (2, i * 2 + 4, '-', 1, percent_bar);
    i++;
  }
  out_hits = total_hits_count - total_hits;
  percent = out_hits * 100;

  sprintf (p, "%02d.%-1d%% : Outside of defined range (over total of %010ld)",
	   percent / total_hits_count, 
           ((percent % total_hits_count) * 10) / total_hits_count, 
           total_hits_count);
  putl (2, i * 2 + 3, cols - 4, p);               
  percent /= 100;
  percent_bar = percent * (cols - 5) / total_hits_count;
  draw (2, i * 2 + 4, '-', 1, percent_bar);

  free (p);

  put_screen (debug_screen_save);
  do {
    i = getykey ();
    if (i == K_Control_L)
    {
      save_screen (1, 1, cols - 1, rows - 1);
      i = 0;
    }
  } while (i == 0);
  free (debug_screen_save);
  debug_screen_save = save;
  put_screen (debug_screen_save);
}
/* ---------------------------------------------------------------------- */
static void
evaluation_dialogue ()
{
  char *save, *p = alloca(cols);
  int key, pos = 1, leave = 0;
  int x1 = (cols / 2) - 20, x2 = cols - x1 - 1;
  int y1 = (rows / 2) - 3 , y2 = rows - y1 - 1;
  int i,len = 0, x_start = 0, y_start = 0, width = x2 - x1 - 1;

  save = debug_screen_save;
  debug_screen_save = get_screen ();

  screen_attr = screen_attr_ffocus;
  double_frame (x1, y1, x2, y2);
  screen_attr = screen_attr_normal;
  for (i = y1 + 1; i < y2; i++)
    putl(x1 + 1,i,x2 - x1 - 2," ");
  do
    {
      screen_attr = screen_attr_normal;

      sprintf(p, "Timer : %s",timer_is_set ? "On" : "Off");

      putl (x1 + 1, y1 + 2, width, p);
      sprintf(p, "Code segments to analyze : %d",total_segments);
      putl (x1 + 1, y1 + 3, width, p);

      screen_attr = (A_green << 4) + A_black;
      sprintf(p, "%s", " ReInit ");
      putl (x1 + 1, y1 + 4, 8, p);
      sprintf(p, "%s", "  View  ");
      putl (x1 + 10, y1 + 4, 8, p);
      sprintf(p, "%s", "  Done  ");
      putl (x1 + 19, y1 + 4, 8, p);
      screen_attr = (A_green << 4) + A_yellow;
      switch (pos)
      {
	case 1:
	  len = 3;
	  x_start = x1 + 9;
	  y_start = y1 + 2;
	  sprintf(p,"%s",timer_is_set ? "On" : "Off");
	  break;
	case 2:
	  len = 8;
	  sprintf(p,"%s"," ReInit ");
	  x_start = x1 + 1;
	  y_start = y1 + 4;
	  break;
	case 3:
	  len = 8;
	  sprintf(p,"%s","  View  ");
	  x_start = x1 + 10;
	  y_start = y1 + 4;
	  break;
	case 4:
	  len = 8;
	  sprintf(p,"%s","  Done  ");
	  x_start = x1 + 19;
	  y_start = y1 + 4;
	  break;
      }
      putl (x_start, y_start, len, p);
      put_screen (debug_screen_save);
      if (!using_com_port_num)
	ScreenSetCursor (y_start, x_start);
      key = getykey();
      switch(key)
      {
	case K_Up:
	case K_EUp:
	case K_BackTab:
	  pos--;
	  if (!pos)
	    pos = 4;
	  break;
	case K_Down:
	case K_EDown:
	case K_Tab:
	  pos++;
	  if (pos > 4)
	    pos = 1;
	  break;
	case K_Return:
	  switch (pos)
	  {
	    case 1:
	      if (timer_is_set)
		reset_timer ();
	      else
		set_timer ();
	      break;
	    case 2:
	    {
	      word32 start_addr = 0, stop_addr = 0;
	      int ok;

	      if (!my_read_string ("Enter start address: "))
		start_addr = parse_expression(4, read_buffer, &ok);
	      if (!my_read_string ("Enter stop address: "))
		stop_addr = parse_expression(4, read_buffer, &ok);
	      init_eval_data (start_addr, stop_addr);
	      break;
	    }
	    case 3:
	      view_eval_data ();
	      break;
	    case 4:
	      leave = 1;
	      break;
	  }
	  break;
      }
    }
  while (!leave);
  free (debug_screen_save);                         
/*  free (p);   */
  debug_screen_save = save;
  put_screen (debug_screen_save);
}
/* ---------------------------------------------------------------------- */
static MENU_ITEM Code_Local_Menu[] = {
       "Current Highlighted Item       ", 1        , 0,
       "Run                          ^R", ITEM_TYPE, 0,
       "Trace into function          ^F", ITEM_TYPE, 0,
       "Step over function           ^O", ITEM_TYPE, 0,
       "Trace to current address     ^T", ITEM_TYPE, 0, 
       "Toggle breakpoint            ^B", ITEM_TYPE, 0,
       "Perfomance Analyzer Settings ^H", ITEM_TYPE, 0,
       "Start Performance Analyzer   ^P", ITEM_TYPE, 0,
       "Display codes at <address>   ^G", ITEM_TYPE, 0,
       "Display codes at <EIP>       ^E", ITEM_TYPE, 0,
       "Set EIP to current address   ^I", ITEM_TYPE, 0,
       "Display 1 byte earlier       ^K", ITEM_TYPE, 0,
       "Display 1 byte later         ^L", ITEM_TYPE, 0,
       0                                , 0        , 0           
};                                                     
/* ---------------------------------------------------------------------- */
static void
code_pane_command (int key)
{
  int b;
            
  if (key == K_Control_X)
  { pop_menu (Code_Local_Menu, 1, 3, &b);
    switch (b)
    { case 1:
        step (2);    /* Run is a global command... */
        break;
      case 2:
        step (0);
        break;
      case 3:    
        step (1);
        break;
      case 4:  
        key = K_Control_T;
        break;
      case 5:
        key = K_Control_B;
        break;
      case 6:
        key = K_Control_H;
        break;
      case 7:
        key = K_Control_P;
        break;
      case 8:
        key = K_Control_G;
        break;
      case 9:
        key = K_Control_E;
        break;
      case 10:
        key = K_Control_I;
        break;
      case 11:
        key = K_Control_K;
        break;
      case 12:
        key = K_Control_L;
        break;
    }  
  }
  switch (key)
    {
    case K_Up:
    case K_EUp:
      if (pane_pos > 0)
	pane_pos--;
      else
	code_pane_goto (code_skip (code_dump_origin, -1));
      break;
    case K_Down:
    case K_EDown:
      if (pane_pos < toplines - 2)
	pane_pos++;
      else
	code_dump_origin =
	  code_pane_pos[0] == code_pane_pos[1]
	    ? code_pane_pos[2] : code_pane_pos[1];
      break;
    case K_PageUp:
    case K_EPageUp:
      code_dump_origin = code_skip (code_dump_origin, 1 - toplines);
      break;
    case K_PageDown:
    case K_EPageDown:
      code_dump_origin = code_dump_last + 1;
      break;
    case K_Control_K:
      code_dump_origin--;
      break;
    case K_Control_L:
      code_dump_origin++;
      break;
    case K_Control_B:
      b = get_breakpoint (BP_Code, 0, code_pane_pos[pane_pos]);
      if (b != -1)
	reset_breakpoint (b);
      else
      {
	b = set_breakpoint (BP_Code, 0, code_pane_pos[pane_pos]);
	edit_breakpoint(b);
      }
      break;
    case K_Control_T:
      b = set_breakpoint (BP_Code, 0, code_pane_pos[pane_pos]);
      step (2);
      reset_breakpoint (b);
      break;
    case K_Control_E:
      code_pane_goto (a_tss.tss_eip);
      break;
    case K_Control_I:
      a_tss.tss_eip = code_pane_pos[pane_pos];
      break;
    case K_Control_G:
      {
	int ok, res;
	res = read_eval (&ok, "");
	if (ok)
	  code_pane_goto (res);
	break;
      }
    case K_Control_H:
      evaluation_dialogue ();
      break;
    case K_Control_P:
      if (timer_is_set)
      {
	start_evaluate ();
	step (2);
	stop_evaluate ();
	reset_timer ();
      } else
	message (CL_Error, "Timer has not been set -- Action aborted.");
      break;
    }
  redraw (0);
}
/* ---------------------------------------------------------------------- */
static void
register_pane_command (int key)
{
  switch (key)
    {
    case K_Left:
    case K_ELeft:
    case K_Up:
    case K_EUp:
      if (pane_pos > 0) pane_pos--;
      break;
    case K_Right:
    case K_ERight:
    case K_Down:
    case K_EDown:
      if (regs_type[pane_pos + 1] && pane_pos < toplines) pane_pos++;
      break;
    default:
      if (key >= ' ' && key < 127)
	{
	  int res, ok;
	  char s[2];

	  s[0] = key; s[1] = '\0';
	  res = read_eval (&ok, key == '=' ? "" : s);
	  if (ok)
	    switch (regs_type[pane_pos])
	      {
	      case 's':
		*(unsigned short *)(regs_addr[pane_pos]) = res & 0xffff;
		break;
	      case 'f':
		a_tss.tss_eflags = (a_tss.tss_eflags & ~0xed5) | (res & 0xed5);
		break;
	      default:
		*regs_addr[pane_pos] = res;
	      }
	}
    }
  redraw (0);
}
/* ---------------------------------------------------------------------- */
static void
flag_pane_command (int key)
{
  static unsigned flagbits[] =
    { 0x0001, 0x0004, 0x0010, 0x0040, 0x0080, 0x0200, 0x0400, 0x0800 };

  switch (key)
    {
    case K_Space:
    case '+':
    case '-':
    case 't':
    case 'T':
      a_tss.tss_eflags ^= flagbits[pane_pos];
      break;
    case '1':
    case 's':
    case 'S':
      a_tss.tss_eflags |= flagbits[pane_pos];
      break;
    case '0':
    case 'r':
    case 'R':
      a_tss.tss_eflags &= ~flagbits[pane_pos];
      break;
    case K_Left:
    case K_ELeft:
    case K_Up:
    case K_EUp:
      if (pane_pos > 0) pane_pos--;
      break;
    case K_Right:
    case K_ERight:
    case K_Down:
    case K_EDown:
      if (pane_pos < toplines && pane_pos < 7) pane_pos++;
      break;
    }
  redraw (0);
}
/* ---------------------------------------------------------------------- */
static void
breakpoint_pane_command (int key)
{
  int b = breakpoint_count ? pane_pos + breakpoint_origin : -1;
  int last = breakpoint_count && (b == breakpoint_count - 1);

  switch (key)
    {
    case K_Delete:
    case K_EDelete:
    case K_Control_D:
      if (b != -1)
	reset_breakpoint (b);
      if (!last)
	break;
      /* else fall through */
    case K_Left:
    case K_ELeft:
    case K_Up:
    case K_EUp:
      if (pane_pos > 0)
	pane_pos--;
      else
	if (breakpoint_origin > 0)
	  breakpoint_origin--;
      break;
    case K_Right:
    case K_ERight:
    case K_Down:
    case K_EDown:
      if (!last)
	if (pane_pos < bottomlines / 2 - 1)
	  pane_pos++;
	else
	  breakpoint_origin++;
      break;
    case K_Control_G:
    case K_Return:
      if (b != -1)
	if (breakpoint_table[b].type == BP_Code)
	  code_pane_goto (breakpoint_table[b].addr);
      break;
    case K_Control_E:
      if (b != -1)
	if (breakpoint_table[b].type == BP_Code)
	  edit_breakpoint(b);
      break;
    }
  redraw (0);
}
/* ---------------------------------------------------------------------- */
static void
read_write_port (void)
{
  char *save, *endp, *p = alloca(cols);
  int key, pos = 1, leave = 0;
  int x1 = 5, x2 = cols - x1 - 1;
  int y1 = (rows / 2) - 5, y2 = rows - y1 - 1;
  int i,len = 0, x_start, y_start, width = x2 - x1 - 13;
  int action = 0, data_size = 1, port_num = 0;
  word32 data = 0;

  if (x2 < x1 || width < 5 || y1 < 0)
    return;

  save = debug_screen_save;
  debug_screen_save = get_screen ();

  screen_attr = screen_attr_ffocus;
  double_frame (x1, y1, x2, y2);
  screen_attr = screen_attr_normal;
  for (i = y1 + 1; i < y2; i++)
    putl(x1 + 1,i,x2 - x1 - 2," ");
  do
    {
      screen_attr = screen_attr_normal;

      sprintf(p, "Action   : %s",
		 action == 0 ? "Input from port" : "Output to port");
      putl (x1 + 1, y1 + 2, width + 11, p);

      sprintf(p, "Size     : %02d (%s)", data_size,
		 (data_size == 1 ? "Byte" :
		   (data_size == 2 ? "Word" : "DWord")));
      putl (x1 + 1, y1 + 3, width + 11, p);

      sprintf(p, "Port     : 0x%04x (%05d)", port_num, port_num);
      putl (x1 + 1, y1 + 4, width + 11, p);
      sprintf(p, "Data     : 0x%08lx", data);
      putl (x1 + 1, y1 + 5, width + 11, p);

      screen_attr = (A_green << 4) + A_black;
      sprintf(p, "%s", "   Ok   ");
      putl (cols / 2 - 10, y1 + 7, 8, p);
      sprintf(p, "%s", " Cancel ");
      putl (cols / 2 + 1, y1 + 7, 8, p);

      screen_attr = (A_green << 4) + A_yellow;
      x_start = x1 + 12;
      y_start = y1 + pos + 1;
      switch (pos)
      {
	case 1:
	  len = width;
	  sprintf (p, "%-*s", width,
		   action == 0 ? "Input from port" : "Output to port");
	  break;
	case 2:
	  len = 2;
	  sprintf(p,"%02d",data_size);
	  break;
	case 3:
	  len = 6;
	  sprintf(p,"0x%04x",port_num);
	  break;
	case 4:
	  len = 10;
	  sprintf(p,"0x%08lx",data);
	  break;
	case 5:
	  len = 8;
	  sprintf(p,"%s","   Ok   ");
	  x_start = cols / 2 - 10;
	  y_start++;
	  break;
	case 6:
	  len = 8;
	  sprintf(p,"%s"," Cancel ");
	  x_start = cols / 2 + 1;
	  break;
      }
      putl (x_start, y_start, len, p);
      put_screen (debug_screen_save);
      if (!using_com_port_num)
	ScreenSetCursor (y_start, x_start);
      key = getykey();
      switch(key)
      {
	case K_Up:
	case K_EUp:
	case K_BackTab:
	  pos--;
	  if (!pos)
	    pos = 6;
	  break;
	case K_Down:
	case K_EDown:
	case K_Tab:
	  pos++;
	  if (pos > 6)
	    pos = 1;
	  break;
	case K_Return:
	  switch (pos)
	  {
	    case 1:
	      if (!my_read_string("Action (i/o): "))
		if (read_buffer[0] == 'o' || read_buffer[0] == 'O')
		  action = 1;
		else
		  action = 0;
	      break;
	    case 2:
	      if (!my_read_string("Size (1,2,4): "))
		switch (read_buffer [0])
		{
		  case '1':
		    data_size = 1;
		    break;
		  case '2':
		    data_size = 2;
		    break;
		  case '4':
		    data_size = 4;
		    break;
		  default :
		    data_size = 1;
		    break;
		}
	      break;
	    case 3:
	      if (!my_read_string("Port number: "))
	      {
		port_num = strtoul(read_buffer, &endp, 0);
		if (port_num && action == 0)  /* Inport */
		  switch (data_size)
		  {
		    case 1:
		      data = inportb (port_num);
		      break;
		    case 2:
		      data = inportw (port_num);
		      break;
		    case 4:
		      data = inportl (port_num);
		      break;
		  }
	      }
	      break;
	    case 4:
	      if (!my_read_string("Data: "))
	      {
		data = strtoul(read_buffer, &endp, 0);
		if (port_num && action == 1)
		  switch (data_size)
		  {
		    case 1:
		      outportb (port_num, data & 0xff);
		      break;
		    case 2:
		      outportw (port_num, data & 0xffff);
		      break;
		    case 4:
		      outportl (port_num, data);
		      break;
		  }
	      }
	      break;
	    case 5:
	    case 6:
	      leave = 1;
	      break;
	  }
	  break;
	case K_Escape:
	  leave = 1;
	  break;
      }
    }
  while (!leave);
  free (debug_screen_save);
  free (p);
  debug_screen_save = save;
  put_screen (debug_screen_save);
}
/* ---------------------------------------------------------------------- */
static MENU_ITEM Data_Local_Menu[] = {
       "Current Highlighted Item       ", 1        , 0,
       "Toggle data write breakpoint ^U", ITEM_TYPE, 0,
       "Toggle read/write breakpoint ^T", ITEM_TYPE, 0,
       "Display codes at <address>   ^G", ITEM_TYPE, 0,
       "Display codes at <EIP>       ^E", ITEM_TYPE, 0,
       "Display codes at <ESP>       ^O", ITEM_TYPE, 0,
       "Display 1 byte earlier       ^K", ITEM_TYPE, 0,
       "Display 1 byte later         ^L", ITEM_TYPE, 0,
       "Display as bytes             ^B", ITEM_TYPE, 0,
       "Display as words             ^W", ITEM_TYPE, 0,
       "Display as dwords            ^D", ITEM_TYPE, 0,
       "Port access                  ^P", ITEM_TYPE, 0,
       "Data structure supports      ^N", ITEM_TYPE, 0,
       0                                , 0        , 0           
};                                        
/* ---------------------------------------------------------------------- */
static void
data_pane_command (int key)
{
  word32 v = data_dump_origin + pane_pos;
  BP_TYPE t;
  int b;
  
  if (key == K_Control_X)
  { pop_menu (Data_Local_Menu, 1, rows - 15, &b);
    switch (b)
    { case 1: 
        key = K_Control_U;
        break;
      case 2:        
        key = K_Control_T;
        break;
      case 3:        
        key = K_Control_G;
        break;
      case 4:        
        key = K_Control_E;
        break;
      case 5:        
        key = K_Control_O;
        break;
      case 6:        
        key = K_Control_K;
        break;
      case 7:        
        key = K_Control_L;
        break;
      case 8:        
        key = K_Control_B;
        break;
      case 9:        
        key = K_Control_W;
        break;
      case 10:        
        key = K_Control_D;
        break;
      case 11:        
        key = K_Control_P;
        break;
      case 12:        
        key = K_Control_N;
        break;
    }
  }
  switch (key)
    {
    case K_Up:
    case K_EUp:
      if (pane_pos >= 8)
	pane_pos -= 8;
      else
	data_dump_origin -= 8;
      break;
    case K_Down:
    case K_EDown:
      if (v + 8 <= data_dump_last)
	pane_pos += 8;
      else
	data_dump_origin += 8;
      break;
    case K_Left:
    case K_ELeft:
      if (pane_pos > 0)
	pane_pos -= data_dump_size;
      else
	data_dump_origin -= 8, pane_pos = 8 - data_dump_size;
      break;
    case K_Right:
    case K_ERight:
      if (v + data_dump_size <= data_dump_last)
	pane_pos += data_dump_size;
      else
	data_dump_origin += 8, pane_pos -= (8 - data_dump_size);
      break;
    case K_PageUp:
    case K_EPageUp:
      data_dump_origin -= (data_dump_last + 1 - data_dump_origin);
      break;
    case K_PageDown:
    case K_EPageDown:
      data_dump_origin = data_dump_last + 1;
      break;
    case K_Control_K:
      data_dump_origin--;
      break;
    case K_Control_L:
      data_dump_origin++;
      break;
    case K_Control_O:
      data_dump_origin = a_tss.tss_esp;
      pane_pos = 0;
      break;
    case K_Control_E:
      data_dump_origin = a_tss.tss_eip;
      data_dump_size = 1;
      pane_pos = 0;
      break;
    case K_Control_B:
      data_dump_size = 1;
      break;
    case K_Control_W:
      data_dump_size = 2;
      break;
    case K_Control_D:
      data_dump_size = 4;
      break;
    case K_Control_N:
      extern_data_support(v);
      break;
    case K_Control_P:
      read_write_port();
      break;
    case K_Control_U:
    case K_Control_T:
      t = key == K_Control_U ? BP_Write : BP_Read;
      b = get_breakpoint (t, data_dump_size, v);
      if (b == -1)
	set_breakpoint (t, data_dump_size, v);
      else
	reset_breakpoint (b);
      break;
    case K_Control_G:
      {
	int ok, res;
	res = read_eval (&ok, "");
	if (ok)
	  {
	    data_dump_origin = res;
	    pane_pos = 0;
	  }
	break;
      }
    default:
      if (key >= ' ' && key < 127)
	{ int res, ok, bad = 0;
	  char s[2], *p, *p0, q;

	  s[0] = key; s[1] = '\0';
	  if (!read_string (key == '=' ? "" : s))
	    { p = read_buffer;
	      do
	      { while (*p == ' ') p++;
		switch (*p)
		{ case '\0':
		    break;
		  case '\'':
		  case '"':
		    if (data_dump_size != 1)
		    { *p = 0;
		      message (CL_Error, "Strings must be entered in byte mode");
		      break;
		    }
		    q = *p++;
		    p0 = p;
		    while (*p != q && *p) 
                      p++;
		    if (*p)
		    { while (p0 != p)
		      { if (valid_addr (v, 1))
			  write_child (v, p0, 1);
			else
			  bad = 1;
			p0++, v++;
		      }
		      if (q == '"')
		      { if (valid_addr (v, 1))
			  write_child (v, s + 1, 1);
			else
			  bad = 1;
			v++;
		      }
		      p++;
		    }
		    else
		      message (CL_Error, "String constant not terminated");
		    break;
		  default:
		    p0 = p;
		    while (*p != ',' && *p)
                      p++;
		    q = *p;
		    *p = 0;
		    res = eval (p0, &ok);
		    if (ok)
		    { *p = q;
		      if (valid_addr (v, data_dump_size))
		        write_child (v, &res, data_dump_size);
		      else
		        bad = 1;
		      v += data_dump_size;
		    }
		}
		if (*p == ',') 
                  p++;
	      } while (*p);
	      if (bad)
		message (CL_Error, "Part of the data could not be written");
	    }
	}
    }
  pane_pos &= ~(data_dump_size - 1);
  redraw (0);
}
/* ---------------------------------------------------------------------- */
static void
npx_pane_command (int key)
{
  int reg = pane_pos - 3, rotreg = 0, tag = 0;
  int regp = (reg >= 0);

  save_npx ();
  if (regp)
    {
      rotreg = (reg + (npx.status >> 11)) & 7;
      tag = (npx.tag >> (rotreg * 2)) & 3;
    }

  switch (key)
    {
    case K_Left:
    case K_ELeft:
    case K_Up:
    case K_EUp:
      if (pane_pos > 0) pane_pos--;
      break;
    case K_Right:
    case K_ERight:
    case K_Down:
    case K_EDown:
      if ((pane_pos < 10) && (pane_pos < toplines - 2)) pane_pos++;
      break;
    case K_Control_C:
    case K_Control_E:
    case K_Delete:
    case K_EDelete:
      if (regp)
	{
	  tag = 3;
	  memset (&npx.reg[reg], 0, sizeof (NPXREG));
	}
      break;
    case K_Control_Z:
      if (regp)
	{
	  tag = 1;
	  memset (&npx.reg[reg], 0, sizeof (NPXREG));
	}
      break;
    case K_Control_N:
      if (regp)
	npx.reg[reg].sign = !npx.reg[reg].sign;
      break;
    default:
      if (regp && key >= ' ' && key < 127)
	{
	  char s[2], *endp, *p;
	  double d;

	  s[0] = key; s[1] = '\0';
	  if (!read_string (key == '=' ? "" : s) && read_buffer[0] != '\0')
	    {
	      p = read_buffer;
	      while (*p == ' ')
		p++;
	      if (*p == '\0')
		break;
	      strlwr (p);
	      if (strcmp (p, "+inf") == 0
		  || strcmp (p, "inf") == 0
		  || strcmp (p, "-inf") == 0)
		{
		  tag = 2;
		  npx.reg[reg].exponent = 0x7fff;
		  npx.reg[reg].sig3 = 0x8000;
		  npx.reg[reg].sig2
		    = npx.reg[reg].sig1
		      = npx.reg[reg].sig0 = 0;
		  npx.reg[reg].sign = (*p == '-');
		}
	      else
		{
		  d = strtod (p, &endp);
		  if (*p != '\0' && *endp)
		    message (CL_Error, "Expression not understood");
		  else
		    {
		      tag = (d == 0.0);
		      *((long double *)(npx.reg + reg)) = (long double) d;
		      npx.reg[reg].sign = (*p == '-'); /* for -Zero */
		    }
		}
	    }
	}
    }
  if (regp)
    npx.tag = (npx.tag & ~(3 << (rotreg * 2))) | (tag << (rotreg * 2));
  load_npx ();
  redraw (0);
}
/* ---------------------------------------------------------------------- */
static void
stack_pane_command (int key)
{
  switch (key)
    {
    case K_Left:
    case K_ELeft:
    case K_Up:
    case K_EUp:
      if (pane_pos > 0)
	pane_pos--;
      else
	if (stack_dump_origin)
	  stack_dump_origin--;
      break;
    case K_Right:
    case K_ERight:
    case K_Down:
    case K_EDown:
      if (pane_pos < stack_dump_last)
	pane_pos++;
      else
	if (stack_dump_more)
	  stack_dump_origin++;
      break;
    case K_Return:
      code_pane_goto (stack_dump_pos[pane_pos]);
      stack_pane_active = 0;
      code_pane_active = 1;
      pane = 0;
      break;
    }
  redraw (0);
}
/* ---------------------------------------------------------------------- */
static void
info_pane_command (int key)
{
  /* No keys recognized.  */
  redraw (0);
}
/* ---------------------------------------------------------------------- */
static void
fullscr_listwild_handler (word32 addr, char typ, char *name,
			  char *filename, int linenum)
{
  char *s;

  s = alloca (cols + strlen (name) + (filename ? strlen (filename) : 0));
  sprintf (s, "0x%08lx %c %s", addr, typ, name);
  if (filename)
    sprintf (s + strlen (s), ", line %d of %s", linenum, filename);
  whereis_pane_text = realloc (whereis_pane_text,
			       (++whereis_text_count) * sizeof (char *));
  whereis_pane_text[whereis_text_count - 1] = strdup (s);
  free (s);
}

static void
whereis_pane_command (int key)
{
  int no = whereis_origin + pane_pos;

  switch (key)
    {
    case K_Left:
    case K_ELeft:
    case K_Up:
    case K_EUp:
      if (pane_pos > 0)
	pane_pos--;
      else
	if (whereis_origin)
	  whereis_origin--;
      break;
    case K_Right:
    case K_ERight:
    case K_Down:
    case K_EDown:
      if (no < whereis_text_count - 1)
	if (pane_pos < toplines - 2)
	  pane_pos++;
	else
	  whereis_origin++;
      break;
    case K_Home:
    case K_EHome:
      pane_pos = whereis_origin = 0;
      break;
    case K_End:
    case K_EEnd:
      if (whereis_text_count > toplines - 1)
	{
	  whereis_origin = whereis_text_count - toplines;
	  pane_pos = toplines - 2;
	}
      else
	{
	  whereis_origin = 0;
	  pane_pos = whereis_text_count ? whereis_text_count - 1 : 0;
	}
      break;
    case K_PageDown:
    case K_EPageDown:
      if (whereis_origin + toplines < whereis_text_count)
	{
	  whereis_origin += toplines - 1;
	  if (whereis_origin + pane_pos >= whereis_text_count)
	    pane_pos = whereis_text_count - 1 - whereis_origin;
	}
      break;
    case K_PageUp:
    case K_EPageUp:
      if (whereis_origin > toplines)
	whereis_origin -= toplines;
      else
	whereis_origin = pane_pos = 0;
      break;
    case K_Return:
      if (whereis_text_count)
	{
	  char *endp, typ;
	  unsigned long ul;

	  typ = whereis_pane_text[no][11];
	  ul = strtoul (whereis_pane_text[no], &endp, 16);
	  switch (toupper (typ))
	    {
	    case 'T':
	      code_pane_goto (ul);
	      whereis_pane_active = 0;
	      code_pane_active = 1;
	      pane = 0;
	      break;
	    case 'B':
	    case 'D':
	      data_dump_origin = ul;
	      pane_positions[4] = 0;
	      break;
	    }
	}
      break;
    default:
      if (key >= ' ' && key < 127)
	{
	  char s[2], *p;

	  s[0] = key; s[1] = '\0';
	  if (!read_string (key == '=' ? "" : s) && read_buffer[0] != '\0')
	    {
	      while (whereis_text_count > 0)
		free (whereis_pane_text[--whereis_text_count]);
	      whereis_pane_text = realloc (whereis_pane_text, 0);

	      p = read_buffer;
	      while (*p == ' ') p++;
	      syms_listwild (p, &fullscr_listwild_handler);
	      pane_pos = whereis_origin = 0;
	      if (whereis_text_count >= 10)
		message (CL_Info, "There were %d symbols that matched",
			 whereis_text_count);
	    }
	}
    }
  redraw (0);
}
/* ---------------------------------------------------------------------- */
static void
module_pane_command(int key)
{
  int no = module_origin + pane_pos;

  switch (key)
  {
    case K_Left:
    case K_ELeft:
    case K_Up:
    case K_EUp:
      if (pane_pos > 0)
	pane_pos--;
      else
	if (module_origin)
	  module_origin--;
      break;
    case K_Right:
    case K_ERight:
    case K_Down:
    case K_EDown:
      if (no < module_text_count - 1)
	if (pane_pos < toplines - 2)
	  pane_pos++;
	else
	  module_origin++;
      break;
    case K_Home:
    case K_EHome:
      pane_pos = module_origin = 0;
      break;
    case K_End:
    case K_EEnd:
      if (module_text_count > toplines - 1)
	{
	  module_origin = module_text_count - toplines;
	  pane_pos = toplines - 2;
	}
      else
	{
	  module_origin = 0;
	  pane_pos = module_text_count ? module_text_count - 1 : 0;
	}
      break;
    case K_PageDown:
    case K_EPageDown:
      if (module_origin + toplines <= module_text_count)
	{
	  module_origin += toplines - 1;
	  if (module_origin + pane_pos >= module_text_count)
	    pane_pos = module_text_count - 1 - module_origin;
	}
      break;
    case K_PageUp:
    case K_EPageUp:
      if (module_origin > toplines - 1)
	module_origin -= toplines - 1;
      else
	module_origin = pane_pos = 0;
      break;
    case K_Return:
      if (no)
        if (!my_read_string ("Line number: "))
        { unsigned long ul;
	  char *endp;

          ul = strtoul (read_buffer, &endp, 0);
          ul = get_linenum (no - 1, ul);
	  code_pane_goto (ul);
	  module_pane_active = 0;
	  code_pane_active = 1;
	  pane = 0;
        }
      break;
    case K_Escape:
      module_pane_active = 0;
      code_pane_active = 1;
      pane = 0;
      break;
  }
  redraw (0);
}
/* ---------------------------------------------------------------------- */
static void
help_pane_command(int key)
{
  int no = help_origin + pane_pos;

  switch (key)
  {
    case K_Left:
    case K_ELeft:
    case K_Up:
    case K_EUp:
      if (pane_pos > 0)
	pane_pos--;
      else
	if (help_origin)
	  help_origin--;
      break;
    case K_Right:
    case K_ERight:
    case K_Down:
    case K_EDown:
      if (no < help_text_count - 1)
	if (pane_pos < toplines - 2)
	  pane_pos++;
	else
	  help_origin++;
      break;
    case K_Home:
    case K_EHome:
      pane_pos = help_origin = 0;
      break;
    case K_End:
    case K_EEnd:
      if (help_text_count > toplines - 1)
	{
	  help_origin = help_text_count - toplines;
	  pane_pos = toplines - 2;
	}
      else
	{
	  help_origin = 0;
	  pane_pos = help_text_count ? help_text_count - 1 : 0;
	}
      break;
    case K_PageDown:
    case K_EPageDown:
      if (help_origin + toplines <= help_text_count)
	{
	  help_origin += toplines;
	  if (help_origin + pane_pos >= help_text_count)
	    pane_pos = help_text_count - 1 - help_origin;
	}
      break;
    case K_PageUp:
    case K_EPageUp:
      if (help_origin > toplines - 1)
	help_origin -= toplines - 1;
      else
	help_origin = pane_pos = 0;
      break;
    case K_Return:
    case K_Escape:
    case K_Space:
      help_pane_active = 0;
      code_pane_active = 1;
      pane = 0;
      break;
  }
  redraw (0);
}
/* ---------------------------------------------------------------------- */
static void
resize_screen(void)
{
  int key, i;

  while (1)
  {
    redraw(-1);
    key = getykey();
    switch (key)
    {
      case K_Up:
      case K_EUp:
	if (toplines > 3)
	{
	  toplines--;
	  bottomlines++;
	}
	break;
      case K_Down:
      case K_EDown:
	if (toplines < rows - 4)
	{
	  toplines++;
	  bottomlines--;
	}
	break;
      case K_Left:
      case K_ELeft:
	switch (pane)
	{
	  case 1:        /* Top middle   */
	  case 2:        /* Top right    */
	    if ((main_x2 > 3) && (main_x2 - main_x1 > 3))
	      main_x2--;
	    break;
	  case 3:        /* Bottom right */
	  case 4:        /* Bottom left  */
	    if (main_x3 > 11 + 2 * data_dump_size)
	      main_x3--;
	    break;
	  default:       /* Top left     */
	    if (main_x1 > 3)
	      main_x1--;
	    break;
	}
	break;
      case K_Right:
      case K_ERight:
	switch (pane)
	{
	  case 1:        /* Top middle   */
	  case 2:        /* Top right    */
	    if (main_x2 < cols - 4)
	      main_x2++;
	    break;
	  case 3:        /* Bottom right */
	  case 4:        /* Bottom left  */
	    if (main_x3 < cols - 4)
	      main_x3++;
	    break;
	  default:       /* Top left     */
	    if ((main_x1 < cols - 4) && (main_x2 - main_x1 > 3))
	      main_x1++;
	    break;
	}
	break;
    }
    if (key == K_Return)
      break;
  }
  for (i = 0; i < PANECOUNT; i++)
    switch(i)
    {
      case 3:        /* Bottom panes  */
      case 4:
	if (pane_positions[i] > bottomlines - 1)
	  pane_positions[i] = bottomlines - 1;
	break;
      default:       /* Top panes     */
	if (pane_positions[i] > toplines - 1)
	  pane_positions[i] = toplines - 1;
	break;
    }
  switch(pane)
  {
    case 3:        /* Bottom panes  */
    case 4:
      if (pane_pos > bottomlines - 1)
	pane_pos = bottomlines - 1;
      break;
    default:       /* Top panes     */
      if (pane_pos > toplines - 1)
	pane_pos = toplines - 1;
      break;
  }
  first_transfer = 1;
  redraw(0);
}
/* ---------------------------------------------------------------------- */
static char *filename = 0;
void
save_screen(int x1, int y1, int x2, int y2)
{
  unsigned char *p;
  int r, c;
  FILE *f;
  struct stat statbuf;

  if (!my_read_string("Save or load? (s/l): "))
  {
    if ((*read_buffer == 'S') || (*read_buffer == 's'))
    {
      p = (unsigned char *)(debug_screen_save + 3);
      if (!filename)
	if (my_read_string("Enter filename: "))
	  return;
	else
	  filename = strdup(read_buffer);

      if (stat(filename,&statbuf) != -1)
	f = fopen(filename,"a+");
      else
	f = fopen(filename,"w+");
      if (f)
      {
	char temp_ch;
	struct date curr_date;
	struct time curr_time;

	getdate (&curr_date);
	gettime (&curr_time);

	for (r = 0; r < 20; r++)
	  fputc ('=', f);
	sprintf (read_buffer, "%02d:%02d:%02d %02d/%02d/%04d ====\n",
		 curr_time.ti_hour, curr_time.ti_min, curr_time.ti_sec,
		 curr_date.da_mon,  curr_date.da_day, curr_date.da_year);
	fputs (read_buffer, f);

	for (r = y1; r < y2; r++)
	{
	  for (c = x1; c < x2; c++)
	  {
	    temp_ch = p[(r * cols + c) * 2];
	    switch (temp_ch)
	    {
	      case 10:
	      case 13:
	      case 26:
		temp_ch = '.';
	    }
	    fputc (temp_ch, f);
	  }
	  fputc ('\n', f);
	}
	fclose(f);
      } else
      {
	message(CL_Error,"Cannot write to file %s",filename);
	free (filename);
	filename = 0;
      }
    } else
    {
      if (!filename)
	if (my_read_string("Enter filename: "))
	  return;
	else
	  filename = strdup(read_buffer);

      f = fopen(filename,"r+");
      if (f)      /* File exist */
      {
	unsigned char *save;
	int not_done = 1, reread = 1;
	long file_row = 0;

	save = debug_screen_save;
	debug_screen_save = get_screen ();
	screen_attr = screen_attr_normal;
	clear_screen ();

	p = malloc (max_cols + 5);
	do {
	  clearerr (f);
	  if (reread)
	  {
	    r = 0;
	    if (fseek (f, file_row * ((long)cols + 1), 0))
	    {
	      clearerr (f);
	      rewind (f);
	    }
	    while ((r < rows) && fgets (p, max_cols + 2, f))
	    {
	      p[strlen(p) - 1] = 0;
	      putl (0, r, cols, p);
	      r++;
	    }
	    if (r < rows)
	      for (; r < rows; r++);
		putl (0, r, cols, " ");
	  }
	  putl (0, rows - 1, cols, "Up/Down/PgUp/PgDn/Enter/Esc");
	  put_screen (debug_screen_save);
	  reread = 1;
	  switch (getykey ())
	  {
	    case K_Return:
	    case K_Escape:
	      not_done = 0;
	      break;
	    case K_Up:
	    case K_EUp:
	      file_row--;
	      if (file_row < 0)
		file_row = 0;
	      break;
	    case K_Down:
	    case K_EDown:
	      file_row++;
	      break;
	    case K_PageUp:
	    case K_EPageUp:
	      file_row -= rows;
	      if (file_row < 0)
		file_row = 0;
	      break;
	    case K_PageDown:
	    case K_EPageDown:
	      file_row += rows;
	      break;
	    default:
	      reread = 0;
	      break;
	  }
	} while (not_done);
	fclose(f);
	free (p);
	free(debug_screen_save);
	debug_screen_save = save;
	put_screen(debug_screen_save);
      } else
      {
	message(CL_Error,"Cannot open file %s",filename);
	free (filename);
	filename = 0;
      }
    }
  }
}
/* ---------------------------------------------------------------------- */
static void
setup_io(void)
{
  if (!my_read_string("Redirect IO to device: "))
  {
    if (using_com_port_num)
      com_done();
    if (strncmp (read_buffer, "con", 3) == 0)
      using_com_port_num = 0;
    if (strncmp (read_buffer, "com_one", 7) == 0)
      using_com_port_num = 1;
    if (strncmp (read_buffer, "com_two", 7) == 0)
      using_com_port_num = 2;
    if (strncmp (read_buffer, "com_three", 9) == 0)
      using_com_port_num = 3;
    if (strncmp (read_buffer, "com_four", 8) == 0)
      using_com_port_num = 4;
    if (using_com_port_num)
    {
      com_init(using_com_port_num);
      first_transfer = 1;
      debug_screen_p = 0;
      if (cols > 80)
         cols = 80;
      if (rows > 24)
        rows = 24;
      toplines = (rows / 2) + 4;
      bottomlines = rows - 3 - toplines;
      main_x1 = cols - 18;
      main_x2 = cols - 5;
      main_x3 = 46;
    }
    free (code_pane_pos);
    free (stack_dump_pos);
    code_pane_pos = malloc ((rows + 2) * sizeof (word32));
    stack_dump_pos = malloc ((rows + 2) * sizeof (word32));
    free (debug_screen_save);
    free (user_screen_save);
    free (transfer_screen_save);
    transfer_screen_save = 0;
    transfer_screen_save = get_screen ();
    transfer_screen_save[1] = transfer_screen_save[2] = 0;
    debug_screen_save = get_screen ();
    debug_screen_save[1] = debug_screen_save[2] = 0;
    user_screen_save = get_screen ();
    user_screen_save[1] = debug_screen_save[2] = 0;
  }
  redraw (0);
}
/* ---------------------------------------------------------------------- */
static void
setup_screen(void)
{
  int new_cols, new_rows;

  if (!my_read_string("Columns: "))
    {
      char *endp;

      new_cols = strtoul(read_buffer, &endp, 10);
      if (new_cols == 0)
	new_cols = cols;
    }
  else
    new_cols = cols;

  if (new_cols < 20)
    new_cols = 20;
  if (new_cols > max_cols)
    new_cols = max_cols;

  if (!my_read_string("Rows: "))
    {
      char *endp;

      new_rows = strtoul(read_buffer, &endp, 10);
      if (new_rows == 0)
	new_rows = rows;
    }
  else
    new_rows = rows;

  if (new_rows < 15)
    new_rows = 15;
  if (new_rows > max_rows)
    new_rows = max_rows;

  cols = new_cols;
  rows = new_rows;

  toplines = (rows / 2) + 4;
  bottomlines = rows - 3 - toplines;
  main_x1 = cols - 18;
  main_x2 = cols - 5;
  main_x3 = 46;
  if (main_x3 > cols)
    main_x3 = cols - 10;

  free (code_pane_pos);
  free (stack_dump_pos);

  code_pane_pos = malloc ((rows + 2) * sizeof (word32));
  stack_dump_pos = malloc ((rows + 2) * sizeof (word32));

  free (debug_screen_save);
  free (transfer_screen_save);
  transfer_screen_save = 0;

  debug_screen_save = get_screen ();
  debug_screen_save[1] = debug_screen_save[2] = 0;  /* Patch cursor pos.  */

  transfer_screen_save = get_screen ();
  transfer_screen_save[1] = transfer_screen_save[2] = 0;

  redraw (0);
}
/* ---------------------------------------------------------------------- */
void
debugger(void)
{
  int oldpane, bar_select, pop_select;
  static void (*keyhandlers[PANECOUNT])(int) =
    {
      &code_pane_command,          /*  0 */
      &register_pane_command,      /*  1 */
      &flag_pane_command,          /*  2 */
      &breakpoint_pane_command,    /*  3 */
      &data_pane_command,          /*  4 */
      &npx_pane_command,           /*  5 */
      &stack_pane_command,         /*  6 */
      &info_pane_command,          /*  7 */
      &whereis_pane_command,       /*  8 */
      &module_pane_command,        /*  9 */
      &help_pane_command           /* 10 */
    };

  main_entry = syms_name2val ("_main");
  first_step = !undefined_symbol;
  initialize ();
  can_longjmp = 1;
  setjmp (debugger_jmpbuf);
  oldpane = -1;
  Main_Menu[0].type = 0;
  display_bar_menu (Main_Menu, &bar_select);
  Main_Menu[0].type = 1;

  while (1)
    {
      int key;

      if (pane < 1 || pane > 4)
	pane = code_pane_active ? 0
	  : (npx_pane_active ? 5
	     : (stack_pane_active ? 6
		: (info_pane_active ? 7
		   : (whereis_pane_active ? 8
		      : (module_pane_active ? 9
			 : (help_pane_active ? 10
			    : (abort (), -1)))))));
      if (pane == oldpane)
	pane_positions[pane] = pane_pos;
      else
	{
	  pane_pos = pane_positions[pane];
	  oldpane = pane;
	  redraw (0);
	}
      key = getykey ();
      switch (key)
	{
        case K_Control_C:
          menu (Main_Menu, &bar_select, &pop_select);
          switch (bar_select)
          {  case 1:       /* File */
               switch (pop_select)
               { case 1:
 	           switch (pane)    /* focused pane */
 	           { case 1:        /* Top middle */
	               save_screen (main_x1 + 1, 2, main_x2, toplines + 1);
	               break;
	             case 2:        /* Top right */
	               save_screen (main_x2 + 1, 2, cols - 1, toplines + 1);
	               break;
	             case 3:        /* Bottom right */
	               save_screen (main_x3+1, toplines+2, cols-1, rows-1);
		       break;
	             case 4:        /* Bottom left */
	               save_screen (1, toplines + 2, main_x3, rows - 1);
	               break;
	             default:       /* Top Left */
	               save_screen (1, 2, main_x1, toplines + 1);
	               break;
	           }
	           break;
                 case 2:
                   setup_screen();
                   break;
                 case 3:    /* Mono mode */
                   screen_attr_normal   = (A_black << 4) + A_grey;
                   screen_attr_source   = screen_attr_normal;
                   screen_attr_focus    = (A_grey  << 4) + A_black;
                   screen_attr_ffocus   = (A_grey  << 4) + A_black;
                   screen_attr_break    = (A_black << 4) + A_white;
                   screen_attr_message  = (A_grey  << 4) + A_white;
                   screen_attr_error    = (A_grey  << 4) + A_black;
                   screen_attr_resize   = (A_grey  << 4) + A_black;
                   screen_attr_asm      = screen_attr_normal;
                   screen_attr_help     = screen_attr_normal;
                   screen_attr_addr     = screen_attr_normal;
                   oldpane = -1;
                   break;
                 case 4:    /* Color */
                   screen_attr_normal   = (A_grey  << 4) + A_blue;
                   screen_attr_source   = (A_grey  << 4) + A_black;
                   screen_attr_focus    = (A_blue  << 4) + A_white;
                   screen_attr_ffocus   = (A_cyan  << 4) + A_white;
                   screen_attr_break    = (A_red   << 4) + A_white;
                   screen_attr_message  = (A_green << 4) + A_white;
                   screen_attr_error    = (A_red   << 4) + A_white;
                   screen_attr_resize   = (A_grey  << 4) + A_green;
                   screen_attr_asm      = (A_grey  << 4) + A_brown;
                   screen_attr_help     = (A_grey  << 4) + A_yellow;
                   screen_attr_addr     = (A_grey  << 4) + A_bold + A_black;
                   oldpane = -1;
                   break;
                 case 5: 
                   setup_io();
                   break;
                 case 6:
	           user_screen ();
	           can_longjmp = 0;
	           if (using_com_port_num)
	             com_done();
	           if (timer_is_set)
	              reset_timer ();
	           return;
               }
               break;
             case 2:       /* Pane */
               switch (pop_select)
               { case 1:
	           if (pane < 1 || pane > 4)
	             pane = 4;
	           else
	             pane--;
	           break;
                 case 2:
	           if (pane < 1 || pane > 4)
	             pane = 1;
	           else
	             pane++;
	           break;
                 case 3:     /* Breakpoint */
                   pane = 4;
                   break;
                 case 5:     /* Data       */
                   pane = 3;
                   break;
                 case 6:     /* Flags      */
                   pane = 2;
                   break;
                 case 11:    /* Registers  */
                   pane = 1;
                   break;
                 case 4:     /* Code       */
                 case 7:     /* Help       */
                 case 8:     /* Info       */
                 case 9:     /* Module     */
                 case 10:    /* Npx        */
                 case 12:    /* Stack      */
                 case 13:    /* Where-is   */
 	           code_pane_active    = (pop_select == 4);
	           help_pane_active    = (pop_select == 7);
	           info_pane_active    = (pop_select == 8);
   	           module_pane_active  = (pop_select == 9);
	           npx_pane_active     = (pop_select == 10);
	           stack_pane_active   = (pop_select == 12);
        	   whereis_pane_active = (pop_select == 13);
                   pane = 0;
                   break;
 	         case 14:
                   resize_screen();
                   pop_select = Main_Menu[0].type;
                   Main_Menu[0].type = 0;
                   display_bar_menu (Main_Menu, &bar_select);
                   Main_Menu[0].type = pop_select;
	           break;
               }
	       oldpane = -1; /* Force redraw.  */
               break;
             case 3:       /* Local */
               keyhandlers[pane] (K_Control_X);
               break;
             case 4:       /* Misc */
               switch (pop_select)
               { case 1:   /* User Screen */
	           user_screen ();
	           (void) getykey ();
	           debug_screen ();
                    break;
                 case 2:   /* Evaluate    */
                 { int res, ok;

	           res = read_eval (&ok, "");
	           if (ok)
	             message (CL_Msg, "\"%s\" -> %d (0x%08lx)",
		              read_buffer, res, (word32) res);
	           break;
	         }
                 case 3:  /* REFRESH */
                   key = K_Control_Z;
                   break;
               }
               break;
             case 5:       /* Help */
               switch (pop_select)
               { case 1:
                    message (CL_Info, 
                      "Sally Full Screen Debugger - Modified by Long Doan");
                    break;
                 case 2:
 	           code_pane_active    = 0;
	           info_pane_active    = 0;
   	           module_pane_active  = 0;
	           npx_pane_active     = 0;
	           stack_pane_active   = 0;
        	   whereis_pane_active = 0;
	           help_pane_active    = 1;
	           oldpane = -1; /* Force redraw.  */
                   break;
               }
               break;
          }          
          break;
	case K_Tab:
	  if (pane < 1 || pane > 4)
	    pane = 1;
	  else
	    pane++;
	  break;
	case K_BackTab:
	  if (pane < 1 || pane > 4)
	    pane = 4;
	  else
	    pane--;
	  break;
	case K_Control_A:
	  if (using_com_port_num)
	  {
	    message (CL_Error, "Local mode only.");
	    break;
	  }
	  if (!my_read_string("Time between step (tens of msecs): "))
	  {
	    unsigned long milisec;
	    char *endp;
	    int stopped = 0;

	    milisec = strtoul(read_buffer, &endp, 10);
	    if (milisec > 0)
	      while (!stopped)
	      {
		step(0);
		stopped = short_sleep(milisec);
	      }
	  }
	  break;
	case K_Control_F:
	  step (0);
	  break;
	case K_Control_O:
	  step (1);
	  break;
	case K_Control_R:
	  step (2);
	  break;         
        case K_Control_Z:
          first_transfer = 1;
          put_screen (debug_screen_save);
          break;
	default:
	  keyhandlers[pane] (key);
	}
    }
}
/* ----------------------------------------------------------------------
   It's a mystery to me --- the game commences
   for the usual fee --- plus expenses
   confidential information --- it's in a diary
   this is my investigation --- it's not a public inquiry

   -- Mark Knopfler, "Private Investigations"
---------------------------------------------------------------------- */
