#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <pc.h>
#include <dos.h>
#include <dpmi.h>
#include <go32.h>
#include <keys.h>
#define far
#include "ed.h"
#include "syms.h"
#include "paging.h"
#include "fsfuncts.h"
#include "xsupport.h"

void dummy_object_support(word32 addr);

/* ---------------------------------------------------------------------- */
static char *
get_old_screen()
{
  char *saved;

  saved = debug_screen_save;
  debug_screen_save = get_screen();
  return saved;
}
/* ---------------------------------------------------------------------- */
static void
restore_screen(char *saved)
{
  free (debug_screen_save);
  debug_screen_save = saved;
}
/* ---------------------------------------------------------------------- */
inline static void
refresh()
{
  put_screen (debug_screen_save);
}
/* ---------------------------------------------------------------------- */
/*
static void
show_buffer()
{
  int y = rows / 2, width = cols - 4;
  char *save;

  save = get_old_screen ();
  screen_attr = screen_attr_message;
  frame (1, y - 1, cols - 2, y + 1);
  putl  (2, y, width, buffer);
  refresh ();
  (void) getykey();
  restore_screen (save);
  refresh ();
}  */
/* ---------------------------------------------------------------------- */
char *
open_win(int width, int height, int attribute, int border,
	 int *x_start, int *y_start, int *actual_width)
{
  char *saved;
  int x1, y1, x2, y2;

  saved = debug_screen_save;
  debug_screen_save = get_screen();
  x1 = cols / 2 - width / 2;
  if (x1 < 1)
    x1 = 1;
  x2 = x1 + width;
  if (x2 > cols - 2)
    x2 = cols - 2;
  y1 = rows / 2 - height / 2;
  if (y1 < 1)
    y1 = 1;
  y2 = y1 + height;
  if (y2 > rows - 2)
    y2 = rows - 2;
  screen_attr = attribute;
  switch (border)
  {
    case 0:
    case 1:
      frame (x1 - 1, y1 - 1, x2, y2);
      break;
    case 2:
      double_frame (x1 - 1, y1 - 1, x2, y2);
  }
  screen_attr = screen_attr_normal;
  *x_start = x1;
  *y_start = y1;
  *actual_width = x2 - x1;
  for (x2 = y1; x2 < y2; x2++)
    putl (x1, x2, *actual_width, " ");
  refresh ();

  return saved;
}
/* ---------------------------------------------------------------------- */
void
close_win(char *saved)
{
  free (debug_screen_save);
  debug_screen_save = saved;
  refresh ();
}
/* ---------------------------------------------------------------------- */
void
init_text (WIN_TEXT *text)
{
  text->page = 0;
  text->lines = 0;
}
/* ---------------------------------------------------------------------- */
void
add_text (WIN_TEXT *text, char *line)
{
  char *p;

  if (!line[0])
    return;

  p = strdup (line);
  text->lines++;
  if (text->lines == 1)
    text->page = malloc (sizeof (char *));
  else
    text->page = realloc (text->page, text->lines * sizeof (char *));
  text->page[text->lines - 1] = p;
}
/* ---------------------------------------------------------------------- */
void
free_text (WIN_TEXT *text)
{
  int i;

  for (i = 0; i < text->lines; i++)
    free (text->page[i]);

  free (text->page);
  text->lines = 0;
  text->page = 0;
}
/* ---------------------------------------------------------------------- */
void
text_win (WIN_TEXT text, int user_width, int user_height)
{
  int x, y, i, width, height, not_done = 1;
  int start_line = 0, pos = 0;
  char *window_save;

  window_save = open_win (user_width, user_height, screen_attr_ffocus,
			  single_border, &x, &y, &width);
  if (user_height > rows - 5)
    height = rows - 5;
  else
    height = user_height;

  while (not_done)
  {
    screen_attr = screen_attr_normal;
    for (i = 0; i < height; i++)
    {
      if (i + start_line < text.lines)
	putl (x, y + i, width, text.page[i + start_line]);
      else
	putl (x, y + i, width , " ");
    }
    screen_attr = screen_attr_focus;
    putl (x, y + pos, width, text.page[pos + start_line]);
    refresh ();
    switch (getykey ())
    {
      case K_Up:
      case K_EUp:
	if (pos == 0)
	{
	  if (start_line)
	    start_line--;
	}
	else
	  pos--;
	break;
      case K_Down:
      case K_EDown:
	if (pos < height - 1)
	{
	  if (start_line + pos < text.lines - 1)
	    pos++;
	}
	else
	{
	  if (start_line + pos < text.lines - 1)
	    start_line++;
	}
	break;
      case K_Shift_F5:
      case K_Control_P:
	save_screen (x, y, x + width, y + height);
	break;
      case K_Return:
      case K_Escape:
	not_done = 0;
    }
  }
  close_win (window_save);
}
/* ---------------------------------------------------------------------- */
static word32
get_sym(char *prompt)
{
  screen_attr = screen_attr_message;
  if (my_read_string(prompt))
    return 0;
  screen_attr = screen_attr_normal;
  return syms_name2val (read_buffer);
}
/* ---------------------------------------------------------------------- */
static int
get_num(char *prompt)
{
  char *endp;

  screen_attr = screen_attr_message;
  if (my_read_string(prompt))
    return 0;
  screen_attr = screen_attr_normal;
  return strtoul (read_buffer, &endp, 0);
}
/* ---------------------------------------------------------------------- */
static int
valid_addr (word32 vaddr, int len)
{
  int a;

  if (vaddr >= 0xffffffffl - len)
    return 0;
  len--;
  for (a = 0; a < MAX_AREA; a++)
    if ((vaddr + len <= areas[a].last_addr) && (vaddr >= areas[a].first_addr))
      return 1;
  if (vaddr >= DOS_LOW_MEM_LO && vaddr <= DOS_LOW_MEM_HI)
    return 1;
  return 0;
}
/* ---------------------------------------------------------------------- */
static word32
get_addr (word32 addr)
{
  word32 return_addr;

  return_addr = get_sym("Address (Enter = Current): ");
  if (!valid_addr (return_addr, sizeof (word32)))
    return_addr = addr;
  return return_addr;
}
/* ---------------------------------------------------------------------- */
int
display_bar_menu (MENU_ITEM menu_data[], int *pos)
{
  int i, len, x = 0;

  for (i = 1; menu_data[i].name && x < cols; i++)
  {
    len = strlen (menu_data[i].name) + 2;
    if (i == menu_data[0].type)
    {
      screen_attr = screen_attr_error;
      *pos = x;
    }
    else
      screen_attr = screen_attr_normal;
    putl (x, 0, len, menu_data[i].name);
    x += len;
  }
  i--;
  screen_attr = screen_attr_normal;
  if (x < cols - 1)
    putl (x, 0, cols - x, " ");
  refresh ();
  return i;
}
/* ---------------------------------------------------------------------- */
int
display_pop_menu (MENU_ITEM menu_data[], int x_pos, int y_pos)
{
  int i, len, y = y_pos + 1, maxlen = 0;

  for (i = 1; menu_data[i].name; i++)
  {
    len = strlen (menu_data[i].name) + 2;
    if (len > maxlen)
      maxlen = len;
  }
  for (i = 1; menu_data[i].name; i++)
  {
    screen_attr = (i == menu_data[0].type 
                   ? screen_attr_error : screen_attr_normal);
    putl (x_pos + 1, y, maxlen, menu_data[i].name);
    y++;
  }
  i--;
  screen_attr = screen_attr_focus;
  frame (x_pos, y_pos, x_pos + maxlen, y);
  screen_attr = screen_attr_normal;
  refresh ();
  return i;
}
/* ---------------------------------------------------------------------- */
void
menu (MENU_ITEM menu_data[], int *BarSelect, int *PopSelect)
{
  int bar_item_num, pop_item_num;
  int not_done, x;
  char *save;

  save = get_old_screen ();

back_to_menu_bar:
  not_done = 1;
  while (not_done)
  {
    bar_item_num = display_bar_menu (menu_data, &x);
    switch (getykey ())
    {
      case K_Left:
      case K_ELeft:
	menu_data[0].type--;
	if (menu_data[0].type < 1)
	  menu_data[0].type = bar_item_num;
	break;
      case K_Right:
      case K_ERight:
	menu_data[0].type++;
	if (menu_data[0].type > bar_item_num)
	  menu_data[0].type = 1;
	break;
      case K_Return:
	not_done = 0;
	break;
      case K_Escape:
	restore_screen (save);
	refresh ();
        *BarSelect = 0;
        *PopSelect = 0;
        return;
    }
  }
  *BarSelect = menu_data[0].type; 
  if (menu_data[*BarSelect].type == ITEM_TYPE)
  {
    restore_screen (save);
    refresh ();
    return;
  }

  not_done = 1;
  while (not_done)
  {
    pop_item_num = display_pop_menu (menu_data[*BarSelect].child, x, 1);
    switch (getykey ())
    {
      case K_Up:
      case K_EUp:
	menu_data[*BarSelect].child[0].type--;
	if (menu_data[*BarSelect].child[0].type < 1)
	  menu_data[*BarSelect].child[0].type = pop_item_num;
	break;
      case K_Down:
      case K_EDown:
	menu_data[*BarSelect].child[0].type++;
	if (menu_data[*BarSelect].child[0].type > pop_item_num)
	  menu_data[*BarSelect].child[0].type = 1;
	break;
      case K_Return:
	not_done = 0;
	break;
      case K_Escape:
	restore_screen (save);
	refresh ();
	save = get_old_screen ();
	goto back_to_menu_bar;
    }
  }
  *PopSelect = menu_data[*BarSelect].child[0].type;
  restore_screen (save);
  refresh ();
}
/* ---------------------------------------------------------------------- */
void
pop_menu (MENU_ITEM menu_data[], int x, int y, int *PopSelect)
{
  int pop_item_num;
  int not_done;
  char *save;

  save = get_old_screen ();

  not_done = 1;
  while (not_done)
  {
    pop_item_num = display_pop_menu (menu_data, x, y);
    switch (getykey ())
    {
      case K_Up:
      case K_EUp:
	menu_data[0].type--;
	if (menu_data[0].type < 1)
	  menu_data[0].type = pop_item_num;
	break;
      case K_Down:
      case K_EDown:
	menu_data[0].type++;
	if (menu_data[0].type > pop_item_num)
	  menu_data[0].type = 1;
	break;
      case K_Return:
	not_done = 0;
	break; 
      case K_Escape:
        restore_screen (save);
        refresh ();
        *PopSelect = 0;
        return;
    }
  }
  *PopSelect = menu_data[0].type;
  restore_screen (save);
  refresh ();
}
/* ---------------------------------------------------------------------- */
void
extern_data_support(word32 addr)
{
  dummy_object_support (addr);
}

/* ---------------------------------------------------------------------- */

/* **********************************************************************
   *                                                                    *
   *       Example for using the xsupport module.                       *
   *                                                                    *
   ********************************************************************** */

static MENU_ITEM dummy_Object_Sub_Menu[] = {
       "Current item          ", 1,         0,
       "Display by index      ", ITEM_TYPE, 0,
       "Display by address    ", ITEM_TYPE, 0,
       "Display hidden objects", ITEM_TYPE, 0,
       "Display object set    ", ITEM_TYPE, 0,
       "Display top instances ", ITEM_TYPE, 0,
       0 	       	       , 0, 0
};
static MENU_ITEM dummy_Frame_Sub_Menu[] = {
       "Current item      ", 1,         0,
       "Display by address", ITEM_TYPE, 0,
       "Display buffer    ", ITEM_TYPE, 0,
       0 	           , 0, 0
};
static MENU_ITEM dummy_Queue_Sub_Menu[] = {
      "Current item            ", 1,         0,
      "Display Queue Descriptor", ITEM_TYPE, 0,
       0       		         , 0, 0
};
static MENU_ITEM dummy_Main_Menu[] = {
       " Current", 1,      0,
       " Objects", MENU_TYPE, dummy_Object_Sub_Menu,
       " Frames ", MENU_TYPE, dummy_Frame_Sub_Menu,
       " Queues ", MENU_TYPE, dummy_Queue_Sub_Menu,
       " Log    ", ITEM_TYPE, 0,
       " Help   ", ITEM_TYPE, 0,
       " Exit   ", ITEM_TYPE, 0,
       0 	         , 0, 0
};

static WIN_TEXT my_win_text;

/* ---------------------------------------------------------------------- */
static void
dummy_disp_help ()
{
  init_text (&my_win_text);
  add_text  (&my_win_text, "Use the <Log> menu or pressing");
  add_text  (&my_win_text, "Shift-F5 or Control-P at any window");
  add_text  (&my_win_text, "to log the current window or");
  add_text  (&my_win_text, "to view the log file.");
  text_win  ( my_win_text, 60, 20);
  free_text (&my_win_text);
}
/* ---------------------------------------------------------------------- */
void
dummy_object_support(word32 addr)
{
  unsigned int index, num;
  word32       new_addr;
  char 	       *old_screen;
  int 	       not_done = 1;
  int	       BarSelect, PopSelect;

  old_screen = get_old_screen ();

  while (not_done)
  {
    screen_attr = screen_attr_message;

    menu (dummy_Main_Menu, &BarSelect, &PopSelect);
    switch (BarSelect)
    {
      case 1:
	switch (PopSelect)
	{
	  case 1: 
	    index = get_num ("Enter an exttype index: ");
	    break;
	  case 2:
	    new_addr  = get_addr (addr);
	    break;
	  case 3:
	    new_addr  = get_addr (addr);
	    break;
	  case 4:
	    new_addr  = get_addr (addr);
	    break;
	  case 5:
	    num = get_num ("Instances: ");
	    break;
	}
	break;
      case 2:
	switch (PopSelect)
	{
	  case 1:
	    new_addr  = get_addr (addr);
	    break;
	  case 2:
	    num = get_num ("Buffer number: ");
	    break;
	}
	break;
      case 3:
	switch (PopSelect)
	{
	  case 1:
	    new_addr  = get_addr (addr);
	    break;
	}
	break;
      case 4:
	if (cols > 31 && rows > 12)
	  save_screen (cols - 31, rows - 31 > 2 ? rows - 31 : 2,
		       cols - 1, rows - 1);
	break;
      case 5:
	dummy_disp_help ();
	break;
      case 6:
	not_done = 0;
	break;
    }
  }

  screen_attr = screen_attr_normal;

  restore_screen (old_screen);
}
